import os
from typing import Optional, List
import click
import panqec
from tqdm import tqdm
import numpy as np
import json
from json.decoder import JSONDecodeError
import multiprocessing
import datetime
import time
import psutil
from .simulation import (
    run_file
)
from .config import CODES, ERROR_MODELS, DECODERS, PANQEC_DIR
from .slurm import (
    get_status, count_input_runs,
    clear_out_folder, clear_sbatch_folder
)
from .utils import (
    get_direction_from_bias_ratio, load_json, save_json, progress_bar
)
from panqec.gui import GUI
from glob import glob
from .usage import summarize_usage
from .analysis import Analysis


@click.group(invoke_without_command=True)
@click.version_option(version=panqec.__version__, prog_name='panqec')
@click.pass_context
def cli(ctx):
    """
    panqec - biased noise in 3D simulations.

    See panqec COMMAND --help for command-specific help.
    """
    if not ctx.invoked_subcommand:
        print(ctx.get_help())


@click.command()
@click.option('-p', '--port', 'port')
def start_gui(port: Optional[int]):
    gui = GUI()
    gui.run(port=port)


@click.command()
@click.pass_context
@click.option('-i', '--input_file', type=str)
@click.option('-o', '--output_file', type=click.STRING)
@click.option('-t', '--trials', default=100, type=click.INT, show_default=True)
def run(
    ctx,
    input_file: Optional[str],
    output_file: str,
    trials: int
):
    """Run a single job or run many jobs from input file."""
    if input_file is not None:
        run_file(
            os.path.abspath(input_file),
            os.path.abspath(output_file),
            trials,
            progress=tqdm
        )
    else:
        print(ctx.get_help())


@click.command()
@click.option('-d', '--data_dir')
@click.option(
    '-t', '--trials', default=1000, type=click.INT, show_default=True
)
@click.option(
    '-n', '--n_nodes', default=1, type=click.INT, show_default=True
)
@click.option(
    '-j', '--job_idx', default=1, type=click.INT, show_default=True
)
@click.option(
    '-c', '--n_cores', default=None, type=click.INT, show_default=True
)
@click.option(
    '--delete-existing', is_flag=True, default=False, show_default=True,
    help="Delete existing results folder in the data directory"
)
def run_parallel(
    data_dir: str,
    trials: int,
    n_nodes: int,
    job_idx: int,
    n_cores: Optional[int],
    delete_existing: bool,
    compressed_output: bool = True
):
    """Run panqec jobs in parallel"""

    input_dir = os.path.join(data_dir, "inputs")
    result_dir = os.path.join(data_dir, "results")
    logs_dir = os.path.join(data_dir, "logs")
    progress_dir = os.path.join(logs_dir, "progress")

    os.makedirs(result_dir, exist_ok=True)
    os.makedirs(progress_dir, exist_ok=True)

    i_node = job_idx - 1

    assert 1 <= job_idx <= n_nodes, \
        f"job_id={job_idx} is invalid. It must be between 1 and {n_nodes}"

    n_cpu = multiprocessing.cpu_count()

    if not n_cores:
        n_cores = n_cpu

    assert n_cores <= n_cpu, \
        f"The number of cores requested ({n_cores}) is higher than" \
        f"the total number of cores ({n_cpu})"

    print(f"Running job {job_idx}/{n_nodes} on {n_cores} cores")

    n_tasks = n_nodes * n_cores

    print(f"Total number of tasks: {n_tasks}\n")

    list_inputs = glob(f"{input_dir}/*.json")

    print("List inputs", list_inputs)

    n_inputs = len(list_inputs)

    if n_inputs == 0:
        raise ValueError(f"No input files in {input_dir}")

    procs = []
    for i_core in range(n_cores):
        i_task = n_cores * i_node + i_core

        n_tasks_per_input = n_tasks // n_inputs

        i_input = i_task // n_tasks_per_input
        if i_input >= n_inputs:
            i_input = n_inputs - 1

        if i_input == n_inputs - 1:
            n_tasks_per_input = n_tasks_per_input + n_tasks % n_inputs

        i_task_in_input = i_task % n_tasks_per_input

        if i_input == n_inputs - 1:
            i_task_in_input = i_task - n_tasks // n_inputs * (n_inputs - 1)

        n_runs = trials // n_tasks_per_input

        if i_task_in_input == n_tasks_per_input - 1:
            n_runs += trials % n_runs

        filename = list_inputs[i_input]
        input_name = os.path.basename(filename)

        # Split the results over files results_1.json, results_2.json, etc.
        max_n_digits = len(str(n_tasks))
        result_json_file = os.path.abspath(os.path.join(
            result_dir,
            f"results_{str(i_task+1).zfill(max_n_digits)}.json"
        ))
        result_gz_file = result_json_file + ".gz"

        for f in [result_json_file, result_gz_file]:
            if delete_existing and os.path.exists(f):
                os.remove(f)

        if compressed_output:
            result_file = result_gz_file
        else:
            result_file = result_json_file

        log_file = os.path.abspath(os.path.join(
            progress_dir,
            f"progress_{str(i_task+1).zfill(max_n_digits)}.txt"
        ))
        print(f"{input_name}\t{n_runs}")

        input_file = os.path.abspath(os.path.join(input_dir, input_name))

        proc = multiprocessing.Process(
            target=run_file,
            args=(input_file, result_file, n_runs),
            kwargs={
                'progress': tqdm,
                'log_file': log_file
            }
        )
        procs.append(proc)
        proc.start()

    # complete the processes
    for proc in procs:
        proc.join()


@click.command()
@click.argument('model_type', required=False, type=click.Choice(
    ['codes', 'error_models', 'decoders'],
    case_sensitive=False
))
def ls(model_type=None):
    """List available codes, error models and decoders."""
    if model_type is None or model_type == 'codes':
        print('Codes:')
        print('\n'.join([
            '    ' + name for name in sorted(CODES.keys())
        ]))
    if model_type is None or model_type == 'error_models':
        print('Error Models (Noise):')
        print('\n'.join([
            '    ' + name for name in sorted(ERROR_MODELS.keys())
        ]))
    if model_type is None or model_type == 'decoders':
        print('Decoders:')
        print('\n'.join([
            '    ' + name for name in sorted(DECODERS.keys())
        ]))


def read_bias_ratios(eta_string: str) -> list:
    """Read bias ratios from comma separated string."""
    bias_ratios = []
    for s in eta_string.split(','):
        s = s.strip()
        if s == 'inf':
            bias_ratios.append(np.inf)
        elif float(s) % 1 == 0:
            bias_ratios.append(int(s))
        else:
            bias_ratios.append(float(s))
    return bias_ratios


def read_range_input(specification: str) -> List[float]:
    """Read range input string and return list."""
    values: List[float] = []
    if ':' in specification:
        parts = specification.split(':')
        min_value = float(parts[0])
        max_value = float(parts[1])
        step = 0.005
        if len(parts) == 3:
            step = float(parts[2])
        values = np.arange(min_value, max_value + step, step).tolist()
    elif ',' in specification:
        values = [float(s) for s in specification.split(',')]
    else:
        values = [float(specification)]
    return values


@click.command()
@click.option(
    '-o', '--overrides', type=click.Path(exists=True),
    default=None,
    help='Overrides specification .json file.'
)
@click.option(
    '-p', '--plot_dir', type=click.Path(),
    default=os.path.join(PANQEC_DIR, 'plots'),
    help='Directory to save plots in.'
)
@click.argument(
    'paths', nargs=-1, type=click.Path(exists=True),
)
def analyze(paths, overrides, plot_dir):
    """Analyze the data at given paths."""

    # Use headless plotting and ignore warnings from matplotlib.
    import matplotlib
    matplotlib.use('Agg')
    import warnings
    warnings.filterwarnings('ignore')

    analysis = Analysis(list(paths), overrides=overrides, verbose=True)
    analysis.analyze(progress=tqdm)
    analysis.make_plots(plot_dir)
    analysis.save(os.path.join(plot_dir, 'analysis.json.gz'))


@click.command()
@click.argument('log_file', type=str, required=True)
@click.option(
    '-i', '--interval', default=10, type=click.INT,
    show_default=True
)
def monitor_usage(log_file: str, interval: float = 10):
    """Continously monitor CPU usage by logging to file at intervals.

    Parameters
    ----------
    log_file : str
        Path to log file where messages are saved.
    interval : int
        Interval at which to check usage, in seconds.
    """
    ppid = os.getppid()
    if not os.path.isfile(log_file):
        with open(log_file, 'w') as f:
            f.write(f'Log file for {ppid}\n')
    while True:
        cpu_usage = psutil.cpu_percent(percpu=True)
        mean_cpu_usage = np.mean(cpu_usage)
        n_cores = len(cpu_usage)
        time_now = datetime.datetime.now()
        mem = psutil.virtual_memory()
        ram_usage = mem.percent
        ram_total = mem.total/2**30
        message = (
            f'{time_now} CPU usage {mean_cpu_usage:.2f}% '
            f'({n_cores} cores) '
            f'RAM {ram_usage:.2f}% ({ram_total:.2f} GiB tot)'
        )
        with open(log_file, 'a') as f:
            f.write(message + '\n')
        time.sleep(interval)


@click.command()
@click.option(
    '-d', '--data_dir', required=True, type=str,
    help='Directory to save input .json files, as'
    '`[data_dir]/inputs/input_bias_[eta].json`'
)
@click.option(
    '--decoder_class', default='BeliefPropagationOSDDecoder',
    show_default=True,
    type=click.Choice(list(DECODERS.keys())),
    help='Decoder class name. '
    'Use `panqec ls decoders` to find the list of all decoders.'
)
@click.option(
    '-s', '--sizes', default='3x3,5x5,7x7', type=str,
    show_default=True,
    help='List of sizes, separated by a comma, where each size'
    'has the form [Lx]x[Ly]x[Lz]'
)
@click.option(
    '--bias', default='Z', type=click.Choice(['X', 'Y', 'Z']),
    show_default=True,
    help='Pauli noise bias'
)
@click.option(
    '--eta', default='0.5', type=str,
    show_default=True,
    help='Bias ratio'
)
@click.option(
    '--prob', default='0:0.6:0.005', type=str,
    show_default=True,
    help='min:max:step or single value or list of values'
)
@click.option(
    '--code_class', default=None, type=str,
    show_default=True,
    help='Code class name, e.g. Toric3DCode. '
    'Use `panqec ls codes` to find the list of all codes'
)
@click.option(
    '--noise_class', default='PauliErrorModel', type=str,
    show_default=True,
    help='Error model class name, e.g. PauliErrorModel. '
    'Use `panqec ls error_models` to find the list of all error models'
)
@click.option(
    '--deformation_name', default=None, type=str,
    show_default=True,
    help='Name of the Clifford deformation to use in our noise, e.g. XZZX'
)
@click.option(
    '-m', '--method', default='direct',
    show_default=True,
    type=click.Choice(['direct', 'splitting']),
    help='Simulation method, between "direct" (simple Monte-Carlo simulation)'
    'and "splitting" (Metropolis-Hastings for low error rates)'
)
@click.option(
    '-l', '--label', default=None,
    show_default=True,
    type=str,
    help='Label for the inputs'
)
def generate_input(
    data_dir, sizes, decoder_class, bias, eta, prob,
    code_class, noise_class, deformation_name, method, label
):
    """Generate the json files of every experiment.

    \b
    Example:
    panqec generate-input -i data/toric-3d-code/ \\
            --code_class Toric3DCode \\
            --noise_class PauliErrorModel \\
            -s 3x3x3,5x5x5,7x7x7, --decoder BeliefPropagationOSDDecoder \\
            --bias Z --eta '10,100,1000,inf' \\
            --prob 0:0.5:0.005
    """
    input_dir = os.path.join(data_dir, 'inputs')
    os.makedirs(input_dir, exist_ok=True)

    error_rates = read_range_input(prob)
    bias_ratios = read_bias_ratios(eta)

    for eta in bias_ratios:
        direction = get_direction_from_bias_ratio(bias, eta)

        L_list = [s.split('x') for s in sizes.split(',')]
        code_parameters = [
            {
                "L_x": int(L[0]),
                "L_y": int(L[1]) if len(L) >= 2 else int(L[0]),
                "L_z": int(L[2]) if len(L) == 3 else int(L[0])
            }
            for L in L_list
        ]
        code_dict = {
            "name": code_class,
            "parameters": code_parameters
        }

        noise_parameters = direction
        if deformation_name is not None:
            noise_parameters['deformation_name'] = deformation_name

        error_model_dict = {
            "name": noise_class,
            "parameters": noise_parameters
        }

        if decoder_class == "BeliefPropagationOSDDecoder":
            decoder_parameters = {'max_bp_iter': 1000,
                                  'osd_order': 100}
        else:
            decoder_parameters = {}

        method_parameters = {}
        if method == 'splitting':
            method_parameters['n_init_runs'] = 20000

        method_dict = {
            'name': method,
            'parameters': method_parameters
        }

        decoder_dict = {"name": decoder_class,
                        "parameters": decoder_parameters}

        if label is None:
            label = 'experiment'

        ranges_dict = {"label": label,
                       "method": method_dict,
                       "code": code_dict,
                       "error_model": error_model_dict,
                       "decoder": decoder_dict,
                       "error_rate": error_rates}

        json_dict = {"comments": "",
                     "ranges": ranges_dict}

        filename = os.path.join(input_dir, f'{label}.json')

        with open(filename, 'w') as json_file:
            json.dump(json_dict, json_file, indent=4)


@click.group(invoke_without_command=True)
@click.pass_context
def slurm(ctx):
    """Routines for generating and running slurm scripts."""
    if not ctx.invoked_subcommand:
        print(ctx.get_help())


@click.command()
@click.argument(
    'result-files', type=click.Path(exists=True), nargs=-1, required=True
)
@click.option(
    '-o', '--output_file', type=str, default='merged-results.json.gz',
    show_default=True
)
def merge_results(
    result_files: str,
    output_file: str = 'merged-results.json.gz'
):
    """Merge result directories that had been split into outdir."""

    print(f'Merging {len(result_files)} files to {output_file}')
    combined_results = []
    for file in result_files:
        try:
            combined_results.append(load_json(file))
        except JSONDecodeError:
            print(f'Error reading {file}, skipping')

    save_json(combined_results, output_file)


@click.command()
@click.argument('header-file', required=True)
@click.option('--output-file', '-o', type=str, required=True)
@click.option('-d', '--data-dir', type=click.Path(exists=True), required=True)
@click.option(
    '--cluster', type=click.Choice(['sge', 'slurm', 'pbs']), required=True
)
@click.option('-n', '--n-nodes', type=click.INT, required=True)
@click.option('-w', '--wall-time', type=str, required=True)
@click.option('-m', '--memory', type=str, required=True)
@click.option(
    '-t', '--trials', type=click.INT, show_default=True, required=True
)
@click.option(
    '-c', '--n-cores', default=None, type=click.INT, show_default=True
)
@click.option('-p', '--partition', default='pml', type=str, show_default=True)
@click.option('-q', '--qos', default='dpart', type=str, show_default=True)
@click.option('--working-dir', default='.', type=str, show_default=True)
@click.option(
    '--delete-existing', is_flag=True, default=False, show_default=True
)
def generate_cluster_script(
    header_file: str,
    output_file: str,
    data_dir: str,
    cluster: str,
    n_nodes: int,
    wall_time: str,
    memory: str,
    trials: int,
    n_cores: Optional[int] = None,
    partition: str = 'pml',
    qos: str = 'dpart',
    working_dir: str = '.',
    delete_existing: bool = False
):
    """Generate a generic cluster script from a given header file"""

    with open(header_file, 'r') as f:
        text = f.read()

    inputs_dir = os.path.join(data_dir, 'inputs')
    assert os.path.isdir(inputs_dir), (
        f'{inputs_dir} missing, please create it and generate inputs'
    )
    log_dir = os.path.join(data_dir, 'logs')
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    name = os.path.basename(data_dir)
    data_dir = os.path.abspath(data_dir)
    delete_option = "--delete-existing" if delete_existing else ""

    i_node_dict = {
        'sge': "$SGE_TASK_ID",
        'slurm': "$SLURM_ARRAY_TASK_ID",
        'pbs': "PBS_ARRAY_INDEX"
    }
    job_id_dict = {
        'sge': '${JOB_ID}',
        'slurm': '${SLURM_JOB_ID}',
        'pbs': '$PBS_JOBID'
    }

    i_node = i_node_dict[cluster]
    job_id = job_id_dict[cluster]

    # If n_cores hasn't been specified, take the maximum number of cores
    if n_cores is None:
        n_cores = multiprocessing.cpu_count()

    replace_map = {
        '${TIME}': wall_time,
        '${MEMORY}': memory,
        '${NAME}': name,
        '${N_NODES}': str(n_nodes),
        '${DATA_DIR}': data_dir,
        '${TRIALS}': str(trials),
        '${N_CORES}': str(n_cores),
        '${QUEUE}': partition,
        '${QOS}': qos,
        '${WORKING_DIR}': working_dir
    }
    for template_string, value in replace_map.items():
        text = text.replace(template_string, value)

    monitor_command = "panqec monitor-usage " \
        f"{log_dir}/usage_{job_id}_{i_node}.txt &"
    run_command = "panqec run-parallel " \
        f"-d {data_dir} " \
        f"-n {n_nodes} " \
        f"-j {i_node} " \
        f"-c {n_cores} " \
        f"-t {trials} " \
        f"{delete_option}"

    with open(output_file, 'w') as f:
        f.write(text + "\n\n")
        f.write(monitor_command + "\n\n")
        f.write(run_command + "\n\n")
        f.write("date")

    print(f'Wrote to {output_file}')


@click.command()
@click.argument('folder', required=True, type=click.Choice(
    ['all', 'out', 'sbatch'],
    case_sensitive=False
))
def clear(folder):
    """Clear generated files."""
    if folder == 'out' or folder == 'all':
        clear_out_folder()
    if folder == 'sbatch' or folder == 'all':
        clear_sbatch_folder()


@click.command()
@click.argument('name', required=True)
def count(name):
    """Count number of input parameters contained."""
    n_runs = count_input_runs(name)
    print(n_runs)


@click.command()
def status():
    """Show the status of running jobs."""
    get_status()


@click.command
@click.argument(
    'data_dirs', type=click.Path(exists=True), nargs=-1,
    required=True
)
def check_usage(data_dirs: str):
    """Check usage of resources."""
    log_dirs = []
    if data_dirs:
        for data_dir in data_dirs:
            log_dir = os.path.join(data_dir, 'logs')
            if not os.path.isdir(log_dir):
                print(f'{log_dir} not a directory')
            else:
                log_dirs.append(log_dir)

    summarize_usage(log_dirs)


@click.command
@click.argument(
    'log_dir', type=click.Path(exists=True), required=True
)
@click.option(
    '-a', '--show-all', is_flag=True, default=False,
    help='Show progress on all the cores individually'
)
def check_progress(log_dir: str, show_all: bool = False):
    """Check usage of resources."""
    if not os.path.isdir(log_dir):
        print(f'{log_dir} not a directory')

    if 'progress' in os.listdir(log_dir):
        progress_dir = os.path.join(log_dir, 'progress')
    else:
        progress_dir = log_dir

    list_files = glob(os.path.join(progress_dir, 'progress_*.txt'))

    total_n = 0
    total_k = 0

    for filename in list_files:
        with open(filename, "r") as f:
            k, n = f.read().split("/")

            total_n += int(n)
            total_k += int(k)

        if show_all:
            progress_bar(int(k), int(n))

    if len(list_files) > 0:
        if show_all:
            print("\n")
        print("Total progress:\n")
        progress_bar(total_k, total_n)


slurm.add_command(status)
slurm.add_command(count)
slurm.add_command(clear)
cli.add_command(start_gui)
cli.add_command(run)
cli.add_command(run_parallel)
cli.add_command(ls)
cli.add_command(slurm)
cli.add_command(generate_input)
cli.add_command(monitor_usage)
cli.add_command(merge_results)
cli.add_command(generate_cluster_script)
cli.add_command(check_usage)
cli.add_command(check_progress)
cli.add_command(analyze)



import sys
import re
import subprocess
import numpy as np


def main():
    usage_txt_list = sys.argv[1:]
    for usage_txt in usage_txt_list:
        y = np.array(get_cpu_usage(usage_txt))
        x = np.arange(len(y))
        plot_points(x, y)


def plot_points(x, y):
    gnuplot = subprocess.Popen(
        ["/usr/bin/gnuplot"],
        stdin=subprocess.PIPE
    )
    gnuplot.stdin.write("set term dumb 79 25\n".encode())
    gnuplot.stdin.write(
        "plot '-' using 1:2 title 'CPU Usage (%)' with linespoints \n".encode()
    )
    for i, j in zip(x, y):
        gnuplot.stdin.write('{} {}\n'.format(i, j).encode())
    gnuplot.stdin.write("e\n".encode())
    gnuplot.stdin.flush()
    gnuplot.stdin.write("\n".encode())


def get_cpu_usage(usage_txt):
    with open(usage_txt) as f:
        lines = f.readlines()
    points = []
    for line in lines:
        match = re.search(r'CPU usage ([\d\.]+)%', line)
        if match:
            points.append(match.group(1))
    return points


if __name__ == '__main__':
    main()



import pytest
from panqec.codes import Planar3DCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestPlanar3DCode(StabilizerCodeTest):

    @pytest.fixture
    def code(self):
        return Planar3DCode(4)



import json
import pytest
import numpy as np
from panqec.bsparse import from_array
from panqec.utils import (
    sizeof_fmt, identity, NumpyEncoder, list_where_str, list_where, set_where,
    format_polynomial, simple_print, find_nearest, get_label
)


@pytest.mark.parametrize(
    'value, expected_result',
    [
        (0, '0.0B'),
        (1024, '1.0KiB'),
        (1024*1024*1.5, '1.5MiB'),
        (1024**8*1.2, '1.2YiB')
    ]
)
def test_sizeof_fmt(value, expected_result):
    assert sizeof_fmt(value) == expected_result


def test_identity():
    x = [1, 2, 3]
    assert x == identity(x)


def test_numpy_encoder():

    data = {
        'foo': np.array([1, 2, 3], dtype=int),
        'bar': np.array([1.2, 2.3, 3.4], dtype=float),
        'more': {
            'baz': np.array([1], dtype=np.uint)[0],
            'float': np.array([1.5], dtype=np.float32)[0],
            'bar': np.array([1.2, 2.3, 3.4], dtype=float),
        },
        'long': np.array([[3, 2, 1], [1, 2, 3]], dtype=int),
        'hello': (1, 2, 3),
    }
    data_str = json.dumps(data, cls=NumpyEncoder)
    new_data = json.loads(data_str)
    assert new_data == {
        'foo': [1, 2, 3],
        'bar': [1.2, 2.3, 3.4],
        'more': {
            'baz': 1,
            'float': 1.5,
            'bar': [1.2, 2.3, 3.4],
        },
        'long': [[3, 2, 1], [1, 2, 3]],
        'hello': [1, 2, 3],
    }


@pytest.mark.parametrize(
    'array, where_str, where_list, where_set',
    [
        (
            np.array([[0, 1], [1, 0]]),
            '01 10',
            [(0, 1), (1, 0)],
            {(0, 1), (1, 0)},
        ),
        (
            np.array([
                [
                    [0, 0, 0],
                    [0, 0, 1],
                    [0, 0, 0],
                ],
                [
                    [0, 0, 0],
                    [1, 0, 0],
                    [0, 0, 0],
                ],
                [
                    [0, 0, 1],
                    [0, 0, 0],
                    [0, 0, 0],
                ],
            ]),
            '012 110 202',
            [(0, 1, 2), (1, 1, 0), (2, 0, 2)],
            {(0, 1, 2), (1, 1, 0), (2, 0, 2)},
        )
    ]
)
def test_list_where_str(array, where_str, where_list, where_set):
    assert list_where_str(array) == where_str
    assert list_where(array) == where_list
    assert set_where(array) == where_set


def test_format_polynomial():
    assert format_polynomial('x', [1, 2, 3], digits=1) == '1.0 + 2.0x + 3.0x^2'
    assert format_polynomial('x', [1.2, 2.1, 3.3], digits=1) == (
        '1.2 + 2.1x + 3.3x^2'
    )
    assert format_polynomial('y', [1.2, -2.1, 3.3], digits=2) == (
        '1.20 - 2.10y + 3.30y^2'
    )
    assert format_polynomial('y', [1.2, 0, 3.3], digits=2) == (
        '1.20 + 3.30y^2'
    )


class TestSimplePrint:

    def test_unint_array(self, capsys):
        a = np.eye(3, dtype=np.uint)
        simple_print(a)
        captured = capsys.readouterr()
        assert captured.out == '100\n010\n001\n'

    def test_sparse_array(self, capsys):
        a = from_array(np.eye(3, dtype=np.uint))
        simple_print(a)
        captured = capsys.readouterr()
        assert captured.out == '100\n010\n001\n'

    def test_sparse_array_space_zeros(self, capsys):
        a = from_array(np.eye(3, dtype=np.uint))
        simple_print(a, zeros=False)
        captured = capsys.readouterr()
        assert captured.out == '1\n 1\n  1\n'

    def test_1d_array(self, capsys):
        a = np.array([1, 0, 1, 0])
        simple_print(a)
        captured = capsys.readouterr()
        assert captured.out == '1010\n'


class TestFindNearest:

    def test_easy_case(self):
        array = [1, 2, 3, 4, 5, 6]
        value = 5.2
        assert find_nearest(array, value) == 5


class TestGetLabel:

    def test_integer_params_only(self):
        label = get_label('Toric2DCode', {'L_x': 5, 'L_y': 6, 'L_z': 7})
        assert label == 'Toric2DCode(L_x=5, L_y=6, L_z=7)'

    def test_float_params_rounded_to_six_digits(self):
        assert get_label('PauliErrorModel', {
            'r_x': 0.33333333333333337,
            'r_y': 0.33333333333333337,
            'r_z': 0.3333333333333333
        }) == 'PauliErrorModel(r_x=0.333333, r_y=0.333333, r_z=0.333333)'



from typing import Tuple, Dict, List
from panqec.codes import StabilizerCode
import numpy as np

Operator = Dict[Tuple, str]  # Location to pauli ('X', 'Y' or 'Z')
Coordinates = List[Tuple]  # List of locations


class Color666ToricCode(StabilizerCode):
    """2D color code on periodic 6,6,6 lattice.

    The overall shape of the lattice is roughly a rhombus.

    Parameters
    ----------
    L_x : int
        Number of unit cells in x direction.
    L_y : Optional[int]
        Number of unit cells in y direction.

    Notes
    -----
    The 6,6,6 lattice is a tessallation of red hexagons (6),
    green hexagons (6) and blue hexagons (6).
    Qubits live on the vertices.
    For each colored shape, there are two stabilizer generators,
    one of all X over the qubits on its corners,
    and one of all Z over the qubits on its corners.

    See `EC zoo article <https://errorcorrectionzoo.org/c/color>`_
    for further reading.
    """
    dimension = 2
    deformation_names = ['X3Z3']

    @property
    def label(self) -> str:
        return 'Color 6.6.6 Toric {}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        coordinates = []

        stab_coordinates = self.get_stabilizer_coordinates()

        for location in stab_coordinates:
            qubit_coords = list(self.get_stabilizer(location).keys())
            for coord in qubit_coords:
                if coord not in coordinates:
                    coordinates.append(coord)

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        for x in range(2, 9*Lx, 3):
            for y in range(2+2*(x-2)//3, 12*Ly+2*(x-2)//3+1, 4):
                coordinates.append((x, y, 0))
                coordinates.append((x, y, 1))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        x, y, pauli = location
        stab_type = 'face'

        if (x % 6 == 2 and y % 12 == 2) or (x % 6 == 5 and y % 12 == 8):
            stab_type += '-green'
        elif (x % 6 == 5 and y % 12 == 0) or (x % 6 == 2 and y % 12 == 6):
            stab_type += '-blue'
        else:
            stab_type += '-red'

        if pauli == 0:
            stab_type += '-x'
        else:
            stab_type += '-z'

        return stab_type

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        Lx, Ly = self.size

        if location[-1] == 0:
            pauli = 'X'
        else:
            pauli = 'Z'

        delta: List[Tuple] = [(-1, -2), (1, -2), (2, 0),
                              (1, 2), (-1, 2), (-2, 0)]

        operator: Operator = dict()
        for d in delta:
            x = (location[0] + d[0]) % (9*Lx)
            y = (location[1] + d[1])

            if (location[0] + d[0] >= 9*Lx):
                y -= 6*Lx

            if (y >= 12*Ly+2*(x-2)//3):
                y -= 12*Ly

            if (x == 0 and y == -2):  # Qubit which was on a corner
                x = 0
                y = 12*Ly - 2

            operator[(x, y)] = pauli

        return operator

    def qubit_axis(self, location: Tuple) -> str:
        x, y = location

        return 'x'

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        **kwargs
    ) -> Dict:

        x, y = location

        if deformation_name == 'X3Z3':
            undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

            x_to_y = {0: 2, 10: 2, 1: 0, 3: 0, 4: 6, 6: 6, 7: 4, 9: 4}

            if x_to_y[x % 12] == y % 8:
                deformation = deformed_dict
            else:
                deformation = undeformed_dict
        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        return deformation

    def get_logicals_x(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        operator: Operator = dict()
        for x in range(8, 9*Lx, 9):
            y = 12*Ly - 6 - 6 * (x-8) // 9
            operator[(x-4, y+4)] = 'X'
            operator[(x-2, y+4)] = 'X'

            if self.is_qubit((x+1, y+2)):
                operator[(x+1, y+2)] = 'X'
                operator[(x+2, y)] = 'X'
            else:
                operator[(0, 2)] = 'X'
                operator[(1, 0)] = 'X'
        logicals.append(operator)

        operator = dict()
        for x in range(5, 9*Lx, 9):
            y = 12*Ly - 4 - 6 * (x-5) // 9
            operator[(x-5, y+2)] = 'X'
            operator[(x-4, y)] = 'X'

            operator[(x-1, y-2)] = 'X'
            operator[(x+1, y-2)] = 'X'
        logicals.append(operator)

        operator = dict()
        for y in range(8, 12*Ly, 12):
            x = 5
            operator[(x-1, y+2)] = 'X'
            operator[(x-2, y)] = 'X'
            operator[(x-2, y-4)] = 'X'
            operator[(x-1, y-6)] = 'X'
        logicals.append(operator)

        operator = dict()
        for y in range(8, 12*Ly, 12):
            x = 5
            operator[(x-2, y)] = 'X'
            operator[(x-1, y-2)] = 'X'
            operator[(x-1, y-6)] = 'X'
            operator[(x-2, y-8)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        operator: Operator = dict()
        for y in range(8, 12*Ly, 12):
            x = 5
            operator[(x-1, y+2)] = 'Z'
            operator[(x-2, y)] = 'Z'
            operator[(x-2, y-4)] = 'Z'
            operator[(x-1, y-6)] = 'Z'
        logicals.append(operator)

        operator = dict()
        for y in range(8, 12*Ly, 12):
            x = 5
            operator[(x-2, y)] = 'Z'
            operator[(x-1, y-2)] = 'Z'
            operator[(x-1, y-6)] = 'Z'
            operator[(x-2, y-8)] = 'Z'
        logicals.append(operator)

        operator = dict()
        for x in range(8, 9*Lx, 9):
            y = 12*Ly - 6 - 6 * (x-8) // 9
            operator[(x-4, y+4)] = 'Z'
            operator[(x-2, y+4)] = 'Z'

            if self.is_qubit((x+1, y+2)):
                operator[(x+1, y+2)] = 'Z'
                operator[(x+2, y)] = 'Z'
            else:
                operator[(0, 2)] = 'Z'
                operator[(1, 0)] = 'Z'
        logicals.append(operator)

        operator = dict()
        for x in range(5, 9*Lx, 9):
            y = 12*Ly - 4 - 6 * (x-5) // 9
            operator[(x-5, y+2)] = 'Z'
            operator[(x-4, y)] = 'Z'

            operator[(x-1, y-2)] = 'Z'
            operator[(x+1, y-2)] = 'Z'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location: Tuple, rotated_picture=False, json_file=None
    ) -> Dict:
        rep = super().stabilizer_representation(location, rotated_picture)

        x, y, _ = location

        # We remove the last part of the location (indexing X or Z)
        rep['location'] = (x, y)

        # Scaling factor for the X hexagons
        if '-x' in self.stabilizer_type(location):
            a = 0.5
        else:
            a = 1

        vertices = np.array(rep['params']['vertices']) * a
        rep['params']['vertices'] = vertices.tolist()

        return rep



from typing import Tuple, Dict, List
import numpy as np
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X','Y','Z')
Coordinates = List[Tuple]  # List of locations


class HollowPlanar3DCode(StabilizerCode):
    dimension = 3

    @property
    def label(self) -> str:
        return 'Hollow Planar 3D {}x{}x{}'.format(*self.size)

    def _is_in_hole(self, x, y, z):
        Lx, Ly, Lz = self.size

        return ((x > 2 and x < 2*Lx-2)
                and (y >= 1 and y < 2*Ly-2)
                and (z >= 1 and z < 2*Lz-2))

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Qubits along e_x
        for x in range(1, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        # Qubits along e_y
        for x in range(2, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(0, 2*Lz, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        # Qubits along e_z
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz-1, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Vertices
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        # Faces in xy plane
        for x in range(1, 2*Lx+1, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(0, 2*Lz, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        # Faces in yz plane
        for x in range(2, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(1, 2*Lz-1, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        # Faces in xz plane
        for x in range(1, 2*Lx+1, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz-1, 2):
                    if not self._is_in_hole(x, y, z):
                        coordinates.append((x, y, z))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        x, y, z = location
        if x % 2 == 0 and y % 2 == 0:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        x, y, z = location

        if self.stabilizer_type(location) == 'vertex':
            delta: List[Tuple] = [(1, 0, 0), (-1, 0, 0), (0, 1, 0),
                                  (0, -1, 0), (0, 0, 1), (0, 0, -1)]
        else:
            # Face in xy-plane.
            if z % 2 == 0:
                delta = [(-1, 0, 0), (1, 0, 0),
                         (0, -1, 0), (0, 1, 0)]
            # Face in yz-plane.
            elif (x % 2 == 0):
                delta = [(0, -1, 0), (0, 1, 0),
                         (0, 0, -1), (0, 0, 1)]
            # Face in zx-plane.
            elif (y % 2 == 0):
                delta = [(-1, 0, 0), (1, 0, 0),
                         (0, 0, -1), (0, 0, 1)]

        operator: Operator = dict()
        for d in delta:
            qubit_location = tuple(np.add(location, d))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location) -> str:
        x, y, z = location

        if (z % 2 == 0) and (x % 2 == 1) and (y % 2 == 0):
            axis = 'x'
        elif (z % 2 == 0) and (x % 2 == 0) and (y % 2 == 1):
            axis = 'y'
        elif (z % 2 == 1) and (x % 2 == 0) and (y % 2 == 0):
            axis = 'z'
        else:
            raise ValueError(f'Location {location} does not correspond'
                             'to a qubit')

        return axis

    def get_logicals_x(self) -> List[Operator]:
        """The unique logical X operator."""

        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for x in range(1, 2*Lx+1, 2):
            operator[(x, 0, 0)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        """The unique logical Z operator."""

        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for y in range(0, 2*Ly, 2):
            for z in range(0, 2*Lz, 2):
                operator[(1, y, z)] = 'Z'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location: Tuple, rotated_picture=False, json_file=None
    ) -> Dict:
        representation = super().stabilizer_representation(location,
                                                           rotated_picture)

        x, y, z = location
        if not rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:  # xy plane
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:  # yz plane
                representation['params']['normal'] = [1, 0, 0]
            else:  # xz plane
                representation['params']['normal'] = [0, 1, 0]

        if rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:
                representation['params']['normal'] = [1, 0, 0]
            else:
                representation['params']['normal'] = [0, 1, 0]

        return representation



"""
Micellaneous useful utilities.

Many of these are copied from the internet.

:Author:
    Eric Huang
"""
from typing import Dict, Any
import numpy as np
import json
import os
import gzip
import hashlib
from panqec.bsparse import is_sparse, to_array
from typing import Callable


def quadratic(x, *params):
    p_th, nu, A, B, C = params
    return A + B*x + C*x**2


def rescale_prob(x_data, *params):
    """Rescaled physical error rate."""
    p, d = x_data
    p_th, nu, A, B, C = params
    x = (p - p_th)*d**nu
    return x


def get_direction_from_bias_ratio(pauli: str, eta) -> dict:
    """Get noise params given Pauli and bias."""

    if eta == np.inf:
        r_bias = 1.
    else:
        r_bias = eta / (1 + eta)
    r_other = (1 - r_bias) / 2

    params: dict = {}

    if pauli == 'Z':
        params = {'r_x': r_other, 'r_y': r_other, 'r_z': r_bias}
    elif pauli == 'X':
        params = {'r_x': r_bias, 'r_y': r_other, 'r_z': r_other}
    elif pauli == 'Y':
        params = {'r_x': r_other, 'r_y': r_bias, 'r_z': r_other}

    return params


def sizeof_fmt(num, suffix='B'):
    """Size to human readable format.

    From stack overflow.
    """
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def nested_map(function: Callable):
    """Return function that applies function to nested lists."""
    def mapper(item):
        if isinstance(item, list):
            return [mapper(x) for x in item]
        else:
            return function(item)

    return mapper


def identity(x):
    return x


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def list_where_str(array):
    return ' '.join(map(
        lambda x: ''.join(map(str, x)),
        sorted(map(tuple, np.array(np.where(array)).T))
    ))


def list_where(array):
    """Get locations of binary list as sorted list of tuples."""
    return sorted(map(tuple, np.array(np.where(array)).T))


def set_where(array):
    """Get locations of binary list as sorted list of tuples."""
    return set(map(tuple, np.array(np.where(array)).T))


def fmt_confidence_interval(
    estimate, left, right, sn=None, unit=None, sn_cutoff=8,
):
    """Format confidence interval for latex."""
    x = estimate
    dx_right = right - estimate
    dx_left = estimate - left
    n_decimals = min(
        -int(np.floor(np.log10(np.abs(dx_left)))),
        -int(np.floor(np.log10(np.abs(dx_right)))),
    )
    leading_magnitude = np.abs(min(dx_left, dx_right))/10**-n_decimals
    if leading_magnitude <= 1.5:
        n_decimals += 1
    if sn is None:
        if np.abs(x) >= 10**sn_cutoff or np.abs(x) <= 10**-sn_cutoff:
            sn = True
        else:
            sn = False
    if sn:
        exponent = int(np.floor(np.log10(np.abs(x))))
        x_mag = np.abs(x)/10**exponent
        dx_mag_left = np.abs(dx_left)/10**exponent
        dx_mag_right = np.abs(dx_right)/10**exponent
    else:
        exponent = 0
        x_mag = np.abs(x)
        dx_mag_left = np.abs(dx_left)
        dx_mag_right = np.abs(dx_right)
    x_round = np.round(x_mag, decimals=n_decimals + exponent)
    dx_left_round = np.round(dx_mag_left, decimals=n_decimals + exponent)
    dx_right_round = np.round(dx_mag_right, decimals=n_decimals + exponent)
    if min(dx_left_round, dx_right_round) > 1.5:
        x_str = str(int(x_round))
    else:
        x_str = str(x_round)
    dx_left_str = str(dx_left_round)
    dx_right_str = str(dx_right_round)

    if sn:
        fmt_str = r'{%s}^{+%s}_{-%s}\times {10}^{%s}' % (
            x_str, dx_right_str, dx_left_str, exponent
        )
        if dx_right_str == dx_left_str:
            fmt_str = r'{(%s\pm %s)}\times {10}^{%s}' % (
                x_str, dx_right_str, exponent
            )
        if x < 0:
            fmt_str = f'-{fmt_str}'
    else:
        fmt_str = r'{%s}^{+%s}_{-%s}' % (x_str, dx_right_str, dx_left_str)
        if dx_left_str == dx_right_str:
            fmt_str = r'{%s}\pm %s' % (x_str, dx_right_str)
    if unit is not None:
        if r'\pm' in fmt_str:
            fmt_str = f'({fmt_str})'
        if unit == r'\%':
            fmt_str += r'\mathrm{%s}' % unit
        else:
            fmt_str += r'\ \mathrm{%s}' % unit
    return f'${fmt_str}$'


def fmt_uncertainty(x, dx, sn=None, sn_cutoff=8, unit=None):
    """Format uncertainty for latex."""
    n_decimals = -int(np.floor(np.log10(np.abs(dx))))
    leading_magnitude = np.abs(dx)/10**-n_decimals
    if leading_magnitude <= 1.5:
        n_decimals += 1
    if sn is None:
        if np.abs(x) >= 10**sn_cutoff or np.abs(x) <= 10**-sn_cutoff:
            sn = True
        else:
            sn = False
    if sn:
        exponent = int(np.floor(np.log10(np.abs(x))))
        x_mag = np.abs(x)/10**exponent
        dx_mag = np.abs(dx)/10**exponent
    else:
        exponent = 0
        x_mag = np.abs(x)
        dx_mag = np.abs(dx)
    x_round = np.round(x_mag, decimals=n_decimals + exponent)
    dx_round = np.round(dx_mag, decimals=n_decimals + exponent)
    if dx_round > 1.5:
        x_str = str(int(x_round))
    else:
        x_str = str(x_round)
    dx_str = str(dx_round)

    if sn:
        fmt_str = r'(%s \pm %s)\times {10}^{%s}' % (
            x_str, dx_str, exponent
        )
        if x < 0:
            fmt_str = f'-{fmt_str}'
    else:
        fmt_str = r'%s \pm %s' % (x_str, dx_str)
        if x < 0:
            fmt_str = f'-({fmt_str})'
    if unit is not None:
        if '(' not in fmt_str:
            fmt_str = f'({fmt_str})'
        fmt_str += r'\ \mathrm{%s}' % unit
    fmt_str = f'${fmt_str}$'
    return fmt_str


def hash_json(dictionary):
    """Produce MD5 hash of dictionary"""
    dict_no_hash = {
        k: v for k, v in dictionary.items()
        if k != 'hash'
    }
    json_string = json.dumps(dict_no_hash, sort_keys=True, indent=2)
    return hashlib.md5(json_string.encode('utf-8')).hexdigest()


def dict_where(signs):
    return set([k for k, v in signs.items() if v])


def face_coords(axis_xyz_list, size):
    """Convert list of (axis, x, y, z) into coordinates."""
    faces = []
    for axis_xyz in axis_xyz_list:
        i, x, y, z = axis_xyz
        lim = 2*np.array(size, dtype=int)
        diff = np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]], dtype=int)
        new = np.mod(2*np.array([x, y, z], dtype=int) + diff[i], lim)
        faces.append(tuple(new))
    return faces


def edge_coords(axis_xyz_list, size):
    """Convert list of (axis, x, y, z) into coordinates."""
    edges = []
    for axis_xyz in axis_xyz_list:
        i, x, y, z = axis_xyz
        lim = 2*np.array(size, dtype=int)
        diff = np.eye(3, dtype=int)
        new = np.mod(2*np.array([x, y, z], dtype=int) + diff[i], lim)
        edges.append(tuple(new))
    return edges


def format_polynomial(symbol, coefficients, digits=3):
    """Polynomial as nice-looking string."""
    poly_str = ''
    format_str = '{:.%df}' % digits
    for power, coefficient in enumerate(coefficients):
        if coefficient < 0:
            sign = '-'
        else:
            sign = '+'
        magnitude = format_str.format(abs(coefficient))
        if power == 0:
            monomial = ''
        elif power == 1:
            monomial = symbol
        else:
            monomial = f'{symbol}^{power}'
        if coefficient != 0:
            if power == 0:
                if sign == '+':
                    poly_str += f'{magnitude}{monomial}'
                else:
                    poly_str += f'{sign}{magnitude}{monomial}'
            else:
                poly_str += f' {sign} {magnitude}{monomial}'
    return poly_str


def simple_print(a, zeros=True):
    """Print binary array in space-saving manner.

    Examples
    --------
    >>> simple_print(np.eye(3))
    100
    010
    001

    >>> simple_print(np.eye(3), zeros=False)
    1
     1
      1
    """
    if is_sparse(a):
        dense = to_array(a)
    else:
        dense = a

    if len(dense.shape) == 1:
        dense = np.array([dense])

    for row in dense:
        row_str = ''
        for entry in row:
            if entry == 0:
                if zeros:
                    row_str += '0'
                else:
                    row_str += ' '
            else:
                row_str += '1'
        print(row_str.rstrip())


def find_nearest(array, value):
    """Find the nearest element in array to given value.

    Parameters
    ----------
    array : Union[List, np.ndarray]
        The array with values in it.
    value : Union[float, int]
        The value to compare to.

    Returns
    -------
    nearest_value : Union[float, int]
        The element in the array that is nearest to the given value.
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def progress_bar(k, n, bar_length=20):
    ratio = k / n
    percentage = str(int(ratio * 100)) + "%"
    n_progress = int(ratio * bar_length)
    n_left = int((1-ratio) * bar_length)

    print(f"{percentage}|" + "█" * n_progress + " " * n_left + f"| {k}/{n}")


def load_json(file):
    if not os.path.isfile(file):
        raise ValueError(f"File {file} not found")
    if os.path.splitext(file)[-1] not in ['.json', '.gz']:
        raise ValueError(f"File {file} is not a .json or .json.gz file")

    if os.path.splitext(file)[-1] == '.json':
        with open(file) as f:
            data = json.load(f)
    elif os.path.splitext(file)[-1] == '.gz':
        with gzip.open(file, 'rb') as gz:
            data = json.loads(gz.read().decode('utf-8'))

    return data


def save_json(data, file):
    if os.path.splitext(file)[-1] not in ['.json', '.gz']:
        raise ValueError(f"Specified file {file} should have extension"
                         ".json or .json.gz")

    if isinstance(data, np.ndarray):
        data = data.tolist()

    if os.path.splitext(file)[-1] == '.json':
        with open(file, 'w') as f:
            json.dump(data, f, cls=NumpyEncoder)
    else:
        with gzip.open(file, 'wb') as gz:
            gz.write(json.dumps(data, cls=NumpyEncoder).encode('utf-8'))


def get_label(name: str, parameters: Dict[str, Any]) -> str:
    """Get label filtering entries from class name and parameters.

    Parameters
    ----------
    name : str
        The name of the class.
    parameters : Dict[str, Any]
        The dictionary of parameters used to intialize an instance of that
        class.

    Returns
    -------
    label : str
        The formatted label.

    Examples
    --------
    >>> get_label('Toric2DCode', {'L_x': 5, 'L_y': 6, 'L_z': 7})
    'Toric2DCode(L_x=5, L_y=6, L_z=7)'
    """
    label = ''
    parameter_labels = []
    for key, value in parameters.items():
        formatted_value = repr(value)
        if isinstance(value, float):
            formatted_value = str(np.round(value, 6))
        parameter_labels.append(f'{key}={formatted_value}')
    label = name + '(' + ', '.join(parameter_labels) + ')'
    return label



from typing import Sequence
import numpy as np
from panqec.codes import StabilizerCode
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel
from typing import Dict
import panqec.bsparse as bsparse
from scipy.sparse import csr_matrix


PAULI_I = 0
PAULI_X = 1
PAULI_Y = 2
PAULI_Z = 3


def symplectic_to_pauli(H):
    n = H.shape[1] // 2
    m = H.shape[0]
    new_H = bsparse.zero_matrix((m, n))
    rows, cols = H.nonzero()
    for i in range(len(rows)):
        if cols[i] < n:
            new_H[rows[i], cols[i]] = PAULI_X
        else:
            if new_H[rows[i], cols[i] - n] == PAULI_X:
                new_H[rows[i], cols[i] - n] = PAULI_Y
            else:
                new_H[rows[i], cols[i] - n] = PAULI_Z
    return new_H


def pauli_to_symplectic(a, reverse=False):
    n = a.shape[0]
    new_a = np.zeros(2*n, dtype='uint8')

    if not reverse:
        new_a[:n] = np.logical_or((a == 1), (a == 2)).astype('uint8')
        new_a[n:] = np.logical_or((a == 2), (a == 3)).astype('uint8')
    else:
        new_a[n:] = np.logical_or((a == 1), (a == 2)).astype('uint8')
        new_a[:n] = np.logical_or((a == 2), (a == 3)).astype('uint8')

    return new_a


def tanh_prod(a, eps=1e-8):
    """ Square cross product defined in II.B of arXiv:2104.13659"""
    prod = np.prod(np.tanh(a/2))
    if prod >= 1:
        prod = 1 - eps
    elif prod <= -1:
        prod = -1 + eps
    return 2 * np.arctanh(prod)


def log_exp_bias(pauli, gamma, eps=1e-12) -> Sequence:
    """ Function lambda defined in II.B of arXiv:2104.13659"""
    denominator = np.sum(np.exp(-gamma), axis=0)
    gamma_pauli = np.choose(pauli, gamma)
    exp_gamma_pauli = np.exp(-gamma_pauli)
    denominator -= exp_gamma_pauli
    return np.log(eps + (1 + exp_gamma_pauli)) - np.log(eps + denominator)


class MemoryBeliefPropagationDecoder(BaseDecoder):
    label = 'MBP decoder'
    allowed_codes = None  # all codes allowed

    def __init__(self,
                 code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 max_bp_iter: int = 100,
                 alpha: float = 0.4,
                 beta: float = 0):
        super().__init__(code, error_model, error_rate)

        self.max_bp_iter = max_bp_iter
        self.alpha = alpha
        self.beta = beta

        self._x_decoder: Dict = dict()
        self._z_decoder: Dict = dict()

        self._decoder: Dict = dict()
        # Convert it to a matrix over GF(4), where each element is in [0,4]
        self.H = code.stabilizer_matrix.toarray()
        self.H_pauli = symplectic_to_pauli(code.stabilizer_matrix).toarray()
        pi, px, py, pz = self.get_probabilities()
        self.p_channel = np.vstack([pi, px, py, pz])

        # Easy access to the neighboring qubits of each stabilizer
        self.neighboring_qubits = [self.H_pauli[m].nonzero()[0]
                                   for m in range(self.H_pauli.shape[0])]

        # Easy access to the neighboring stabilizers of each qubit
        self.neighboring_stabs = [self.H_pauli[:, n].nonzero()[0]
                                  for n in range(self.H_pauli.shape[1])]

        # ===================== Initialize BP variables ====================

        # Create channel log ratios
        self.lambda_channel = np.log((1 - self.p_channel[1:])
                                     / self.p_channel[1:])

        # Initialize [qubit to stabilizer] messages (gamma)
        n_stabs, n_qubits = self.H_pauli.shape
        self.gamma_q2s = np.zeros((3, n_qubits, n_stabs))
        for n in range(n_qubits):
            for m in self.neighboring_stabs[n]:
                for w in range(3):
                    if 1 + w != self.H_pauli[m, n]:
                        self.gamma_q2s[w, n, m] = self.lambda_channel[w, n]

        # Initialize [stabilizer to qubit] messages (delta)
        self.delta_s2q = np.zeros((n_stabs, n_qubits))

    @property
    def params(self) -> dict:
        return {
            'max_bp_iter': self.max_bp_iter,
            'alpha': self.alpha,
            'beta': self.beta
        }

    def get_probabilities(self):
        error_rate = 0.5
        pi, px, py, pz = self.error_model.probability_distribution(
            self.code,
            error_rate
        )
        return pi, px, py, pz

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X and Z corrections given code and measured syndrome."""

        H_pauli = self.H_pauli

        n_stabs, n_qubits = H_pauli.shape

        # ==================== Initialize BP variables ====================

        gamma_q2s = self.gamma_q2s.copy()
        delta_s2q = self.delta_s2q.copy()

        # ========================= BP iterations =========================

        for iter in range(self.max_bp_iter):
            print(f"\nIter {iter+1} / {self.max_bp_iter}")

            gamma_q = np.zeros((3, n_qubits))
            for n in range(n_qubits):
                # --------- Stabilizer to qubit update (prod-sum) ---------

                for m in self.neighboring_stabs[n]:
                    lambda_neighbor = np.array([
                        log_exp_bias(
                            H_pauli[m, n_prime]-1, gamma_q2s[:, n_prime, m]
                        )
                        for n_prime in self.neighboring_qubits[m]
                        if n_prime != n
                    ])
                    lambda_neighbor = np.array(lambda_neighbor)

                    sign = (-1)**syndrome[m]
                    delta_s2q[m, n] = sign * tanh_prod(lambda_neighbor)

                # ----------------- Qubit to stabilizer update ---------------

                for w in range(3):
                    sum_same_pauli = np.sum(
                        delta_s2q[self.neighboring_stabs[n], n][
                            H_pauli[self.neighboring_stabs[n], n] == w + 1
                        ]
                    )
                    sum_diff_pauli = np.sum(
                        delta_s2q[self.neighboring_stabs[n], n][
                            H_pauli[self.neighboring_stabs[n], n] != w + 1
                        ]
                    )

                    gamma_q[w, n] = self.lambda_channel[w, n] \
                        + 1 / self.alpha * sum_diff_pauli \
                        - self.beta * sum_same_pauli

                    # Update qubit to stab messages
                    gamma_q2s[w, n, :] = gamma_q[w, n]

                    # Inhibition loop
                    gamma_q2s[w, n, (1 + w != H_pauli[:, n])] -= delta_s2q[
                        (1 + w != H_pauli[:, n]), n
                    ]

            # ----------------------- Hard decision -----------------------

            correction = np.zeros(n_qubits, dtype='uint8')

            for n in range(n_qubits):
                if not np.all(gamma_q[:, n] > 0):
                    correction[n] = np.argmin(gamma_q[:, n]) + 1

            correction_symplectic = pauli_to_symplectic(correction)

            # --------------- Break loop if syndrome reached ---------------

            new_syndrome = self.code.measure_syndrome(correction_symplectic)
            if np.all(new_syndrome == syndrome):
                # print(f"Syndrome reached in {iter} iterations\n")
                break

        correction_symplectic = pauli_to_symplectic(correction, reverse=True)

        correction = np.concatenate([correction_symplectic[n_qubits:],
                                     correction_symplectic[:n_qubits]])

        return correction


def test_symplectic_to_pauli():
    H = csr_matrix(np.array([[1, 1, 0, 1], [1, 0, 1, 1]]))
    print(symplectic_to_pauli(H))


def test_decoder():
    from panqec.codes import Toric3DCode
    from panqec.error_models import PauliErrorModel
    import time

    L = 4
    max_bp_iter = 10
    alpha = 0.4

    code = Toric3DCode(L)

    error_rate = 0.3
    r_x, r_y, r_z = [0.333, 0.333, 0.334]
    error_model = PauliErrorModel(r_x, r_y, r_z)

    decoder = MemoryBeliefPropagationDecoder(
        code, error_model, error_rate, max_bp_iter=max_bp_iter, alpha=alpha
    )

    # Start timer
    start = time.time()

    n_iter = 1
    for i in range(n_iter):
        print(f"\n\nRun {code.label} {i}...")
        print("Generate errors")
        error = np.zeros(2*code.n)
        error[10] = 1
        error[52] = 1
        error[58] = 1

        print("Calculate syndrome")
        syndrome = code.measure_syndrome(error)

        print("Decode")
        correction = decoder.decode(syndrome)
        print("Get total error")
        total_error = (correction + error) % 2
        print("Get effective error")
        effective_error = code.logical_errors(total_error)

        print("Check codespace")
        codespace = code.in_codespace(total_error)
        success = bool(np.all(effective_error == 0)) and codespace

        print("Success:", success)

    print("Average time per iteration", (time.time() - start) / n_iter)


if __name__ == '__main__':
    test_decoder()



from typing import Tuple, Dict
import numpy as np
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]


class SweepDecoder3D(BaseDecoder):

    label: str = 'Toric 3D Sweep Decoder'
    allowed_codes = ["Toric3DCode", "Planar3DCode"]

    _rng: np.random.Generator
    max_sweep_factor: int

    def __init__(self,
                 code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 seed: int = 0,
                 max_sweep_factor: int = 32):
        super().__init__(code, error_model, error_rate)
        self._rng = np.random.default_rng(seed)
        self.max_sweep_factor = max_sweep_factor
        self.seed = seed

    @property
    def params(self) -> dict:
        return {
            'seed': self.seed,
            'max_sweep_factor': self.max_sweep_factor
        }

    def get_face_syndromes(
        self, full_syndrome: np.ndarray
    ) -> np.ndarray:
        """Get only the syndromes for the vertex Z stabilizers.
        Z vertex stabilizers syndromes are discarded for this decoder.
        """
        face_syndromes = self.code.extract_x_syndrome(full_syndrome)
        return face_syndromes

    def flip_edge(
        self, location: Tuple, signs: np.ndarray
    ):
        """Flip signs at index and update correction."""
        x, y, z = location
        edge = tuple(np.mod(location, 2))
        L_x, L_y, L_z = self.code.size
        limits = (2*L_x, 2*L_y, 2*L_z)

        if edge == (1, 0, 0):
            face_1 = (x, y + 1, z)
            face_2 = (x, y - 1, z)
            face_3 = (x, y, z + 1)
            face_4 = (x, y, z - 1)
        elif edge == (0, 1, 0):
            face_1 = (x, y, z + 1)
            face_2 = (x, y, z - 1)
            face_3 = (x + 1, y, z)
            face_4 = (x - 1, y, z)
        elif edge == (0, 0, 1):
            face_1 = (x + 1, y, z)
            face_2 = (x - 1, y, z)
            face_3 = (x, y + 1, z)
            face_4 = (x, y - 1, z)

        # Impose periodic boundary conditions.
        location_1 = tuple(np.mod(face_1, limits))
        location_2 = tuple(np.mod(face_2, limits))
        location_3 = tuple(np.mod(face_3, limits))
        location_4 = tuple(np.mod(face_4, limits))

        # Flip the signs (well actually 0s and 1s).
        if self.code.is_stabilizer(location_1):
            signs[self.code.stabilizer_index[location_1]] = 1 - signs[self.code.stabilizer_index[location_1]]  # type: ignore # noqa
        if self.code.is_stabilizer(location_2):
            signs[self.code.stabilizer_index[location_2]] = 1 - signs[self.code.stabilizer_index[location_2]]  # type: ignore # noqa
        if self.code.is_stabilizer(location_3):
            signs[self.code.stabilizer_index[location_3]] = 1 - signs[self.code.stabilizer_index[location_3]]  # type: ignore # noqa
        if self.code.is_stabilizer(location_4):
            signs[self.code.stabilizer_index[location_4]] = 1 - signs[self.code.stabilizer_index[location_4]]  # type: ignore # noqa

    def get_default_direction(self):
        """The default direction when all faces are excited."""
        direction = int(self._rng.choice([0, 1, 2], size=1))
        return direction

    def get_initial_state(self, syndrome: np.ndarray) -> np.ndarray:
        """Get initial cellular automaton state from syndrome."""
        signs = syndrome.copy()
        signs[self.code.z_indices] = 0

        return signs

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get Z corrections given measured syndrome."""

        # Maximum number of times to sweep before giving up.
        max_sweeps = self.max_sweep_factor*int(max(self.code.size))

        # The syndromes represented as an array of 0s and 1s.
        signs = self.get_initial_state(syndrome)

        # Keep track of the correction needed.
        correction: Dict = dict()

        # Initialize the number of sweeps.
        i_sweep = 0

        # Keep sweeping until there are no syndromes.
        while any(signs) and i_sweep < max_sweeps:
            signs = self.sweep_move(signs, correction)
            i_sweep += 1

        return self.code.to_bsf(correction)

    def sweep_move(
        self, signs: np.ndarray, correction: Operator
    ) -> np.ndarray:
        """Apply the sweep move once."""

        new_signs = signs.copy()

        flip_locations = []

        L_x, L_y, L_z = self.code.size
        limits = (2*L_x, 2*L_y, 2*L_z)

        # Sweep through every edge.
        for x, y, z in np.array(self.code.stabilizer_coordinates)[self.code.z_indices]:  # type: ignore  # noqa: E501
            # Get the syndromes on each face in sweep direction.
            x_face_loc = tuple(np.mod((x, y + 1, z + 1), limits))
            y_face_loc = tuple(np.mod((x + 1, y, z + 1), limits))
            z_face_loc = tuple(np.mod((x + 1, y + 1, z), limits))
            x_face = self.code.is_stabilizer(x_face_loc) and signs[self.code.stabilizer_index[x_face_loc]]  # type: ignore  # noqa: E501
            y_face = self.code.is_stabilizer(y_face_loc) and signs[self.code.stabilizer_index[y_face_loc]]  # type: ignore  # noqa: E501
            z_face = self.code.is_stabilizer(z_face_loc) and signs[self.code.stabilizer_index[z_face_loc]]  # type: ignore  # noqa: E501

            x_edge = tuple(np.mod((x + 1, y, z), limits))
            y_edge = tuple(np.mod((x, y + 1, z), limits))
            z_edge = tuple(np.mod((x, y, z + 1), limits))

            if x_face and y_face and z_face:
                direction = self.get_default_direction()
                flip_locations.append(
                    {0: x_edge, 1: y_edge, 2: z_edge}[direction]
                )
            elif y_face and z_face:
                flip_locations.append(x_edge)
            elif x_face and z_face:
                flip_locations.append(y_edge)
            elif x_face and y_face:
                flip_locations.append(z_edge)

        for location in flip_locations:
            self.flip_edge(location, new_signs)
            correction[location] = 'Z'

        return new_signs



import pytest
from panqec.codes import RhombicToricCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestRhombicToricCode(StabilizerCodeTest):

    @pytest.fixture
    def code(self):
        return RhombicToricCode(6, 6, 6)

    def test_stabilizer_index(self, code):
        assert all(
            len(index) in [3, 4] for index in code.stabilizer_index
        )



import pytest
from typing import Tuple, List
from abc import ABCMeta, abstractmethod
import numpy as np
from scipy.special import comb
from scipy.sparse import find
from tqdm import tqdm
from itertools import combinations
from panqec.codes import StabilizerCode
from panqec.bsparse import is_sparse, vstack, to_array
from panqec.bpauli import bs_prod, brank, bvector_to_pauli_string


class StabilizerCodeTest(metaclass=ABCMeta):

    @pytest.fixture
    @abstractmethod
    def code(self) -> StabilizerCode:
        pass

    def test_n_equals_len_qubit_index(self, code):
        assert code.n == len(code.qubit_coordinates)

    def test_len_vertex_index_equals_number_of_vertex_stabilizers(self, code):
        if 'vertex' in code.stabilizer_types and code.is_css:
            n_vertices = len(code.type_index('vertex'))
            assert n_vertices == code.Hz.shape[0]

    def test_len_face_index_equals_number_of_face_stabilizers(self, code):
        if 'face' in code.stabilizer_types and code.is_css:
            n_faces = len(code.type_index('face'))
            assert n_faces == code.Hx.shape[0]

    def test_qubit_stabilizer_indices_no_overlap(self, code):
        qubits = set(code.qubit_index.keys())
        stabilizers = set(code.stabilizer_index.keys())
        assert qubits.isdisjoint(stabilizers)

    def test_all_stabilizers_commute(self, code):
        commutators = bs_prod(code.stabilizer_matrix, code.stabilizer_matrix)
        print_non_commuting(
            code, commutators, code.stabilizer_matrix, code.stabilizer_matrix,
            'stabilizer', 'stabilizer'
        )

        # There should be no non-commuting pairs of stabilizers.
        assert np.all(commutators == 0)

    def test_logicals_same_size(self, code):
        assert len(code.logicals_x) == len(code.logicals_z)

    def test_logical_operators_anticommute_pairwise(self, code):
        k = code.k
        assert np.all(bs_prod(code.logicals_x, code.logicals_x) == 0)
        assert np.all(bs_prod(code.logicals_z, code.logicals_z) == 0)
        commutators = bs_prod(code.logicals_x, code.logicals_z)
        assert np.all(commutators == np.eye(k)), (
            f'Not pairwise anticommuting {commutators}'
        )

    def test_logical_operators_commute_with_stabilizers(self, code):
        x_commutators = bs_prod(code.logicals_x, code.stabilizer_matrix)
        print_non_commuting(
            code, x_commutators, code.logicals_x, code.stabilizer_matrix,
            'logicalX', 'stabilizer'
        )
        z_commutators = bs_prod(code.logicals_z, code.stabilizer_matrix)
        print_non_commuting(
            code, z_commutators, code.logicals_z, code.stabilizer_matrix,
            'logicalZ', 'stabilizer'
        )
        assert np.all(to_array(x_commutators) == 0), (
            'logicalX not commuting with stabilizers'
        )
        assert np.all(to_array(z_commutators) == 0), (
            'logicalZ not commuting with stabilizers'
        )

    def test_logical_operators_are_independent_by_rank(self, code):
        n, k = code.n, code.k
        matrix = code.stabilizer_matrix

        matrix_with_logicals = vstack([
            matrix,
            code.logicals_x,
            code.logicals_z,
        ])

        rank_with_logicals = brank(matrix_with_logicals)
        assert rank_with_logicals == n + k

    def test_n_indepdent_stabilizers_equals_n_minus_k(self, code):
        n, k = code.n, code.k
        matrix = code.stabilizer_matrix

        # Number of independent stabilizer generators.
        rank = brank(matrix)

        assert rank <= matrix.shape[0]
        assert rank == n - k

    def test_number_of_logicals_is_k(self, code):
        k = code.k
        assert code.logicals_x.shape[0] == k
        assert code.logicals_z.shape[0] == k


class StabilizerCodeTestWithCoordinates(StabilizerCodeTest, metaclass=ABCMeta):

    @property
    @abstractmethod
    def size(self) -> Tuple[int, int, int]:
        """Plane size x"""

    @property
    @abstractmethod
    def code_class(self):
        """The code Class to be tested"""

    @property
    @abstractmethod
    def expected_plane_edges_xy(self) -> List[Tuple[int, int]]:
        """Expected xy coordinates of xy plane edges in unrotated lattice."""

    @property
    @abstractmethod
    def expected_plane_faces_xy(self) -> List[Tuple[int, int]]:
        """Expected xy coordinates of xy plane faces in unrotated lattice."""

    @property
    @abstractmethod
    def expected_vertical_faces_xy(self) -> List[Tuple[int, int]]:
        """Expected xy coordinates of xy plane faces in unrotated lattice."""

    @property
    @abstractmethod
    def expected_plane_vertices_xy(self) -> List[Tuple[int, int]]:
        """Expected xy coordinates of xy plane vertices in unrot lattice."""

    @property
    @abstractmethod
    def expected_plane_z(self) -> List[int]:
        """Expected z coordinates of horizontal planes."""

    @property
    @abstractmethod
    def expected_vertical_z(self) -> List[int]:
        """Expected z coordinates of vertical edges and faces."""

    @pytest.fixture
    def code(self):
        return self.code_class(*self.size)

    def test_qubit_indices(self, code):
        locations = set(code.qubit_coordinates)
        expected_locations = set([
            (x, y, z)
            for x, y in self.expected_plane_edges_xy
            for z in self.expected_plane_z
        ] + [
            (x, y, z)
            for x, y in self.expected_plane_vertices_xy
            for z in self.expected_vertical_z
        ])
        assert locations == expected_locations

    def test_vertex_indices(self, code):
        locations = set(code.type_index('vertex').keys())
        expected_locations = set([
            (x, y, z)
            for x, y in self.expected_plane_vertices_xy
            for z in self.expected_plane_z
        ])
        assert locations == expected_locations

    def test_face_indices(self, code):
        locations = set(code.type_index('face').keys())
        expected_locations = set([
            (x, y, z)
            for x, y in self.expected_vertical_faces_xy
            for z in self.expected_vertical_z
        ] + [
            (x, y, z)
            for x, y in self.expected_plane_faces_xy
            for z in self.expected_plane_z
        ])
        assert locations == expected_locations

    @pytest.mark.skip(reason='brute force')
    def test_find_lowest_weight_Z_only_logical_by_brute_force(self, code):
        n, k = code.n, code.k
        matrix = code.stabilizer_matrix

        coords = {v: k for v, k in enumerate(code.qubit_coordinates)}

        min_weight = 4
        max_weight = 4
        for weight in range(min_weight, max_weight + 1):
            n_comb = comb(n, weight, exact=True)
            for sites in tqdm(combinations(range(n), weight), total=n_comb):
                logical = np.zeros(2*n, dtype=np.uint)
                for site in sites:
                    x, y, z = coords[site]
                    deform = False
                    if z % 2 == 1 and (x + y) % 4 == 2:
                        deform = True

                    if deform:
                        logical[site] = 1  # X operator on deformed
                    else:
                        logical[n + site] = 1  # Z operator on undeformed

                codespace = np.all(bs_prod(matrix, logical) == 0)
                if codespace:
                    matrix_with_logical = np.concatenate([
                        to_array(matrix),
                        [logical]
                    ])
                    if brank(matrix_with_logical) == n - k + 1:
                        print('Found Z-only logical')
                        print([coords[site] for site in sites])
                        print(operator_spec(code, logical))
                        return


def operator_spec(code, bsf):
    """Get representation of BSF as list of (pauli, (x, y, z)) entries.

    Useful for debugging and reading BSF in human-readable format.
    """
    operator_spec = []
    if len(bsf.shape) != 1:
        bsf_1d = to_array(bsf)[0]
    else:
        bsf_1d = to_array(bsf)
    pauli_string = bvector_to_pauli_string(bsf_1d)
    for index, pauli in enumerate(pauli_string):
        if pauli != 'I':
            matches = [
                xyz
                for i, xyz in enumerate(code.qubit_coordinates)
                if i == index
            ]
            location = matches[0]
            operator_spec.append((pauli, location))
    return operator_spec


def print_non_commuting(
    code: StabilizerCode, commutators: np.ndarray,
    operators_1: np.ndarray, operators_2: np.ndarray,
    name_1: str, name_2: str,
    max_print: int = 5
):
    if is_sparse(commutators):
        non_commuting = set([
            (i, j)
            for i, j in np.array(find(commutators))[:2].T
            if i <= j
        ])
    else:
        non_commuting = set([
            (i, j)
            for i, j in np.array(np.where(commutators)).T
            if i <= j
        ])

    # Print the first few non-commuting stabilizers if any found.
    if non_commuting:
        print(
            f'{len(non_commuting)} pairs of non-commuting '
            f'{name_1} and {name_2}:'
        )
        for i_print, (i, j) in enumerate(non_commuting):
            try:
                loc_1 = code.stabilizer_coordinates[i]
            except IndexError:
                loc_1 = ''
            try:
                loc_2 = code.stabilizer_coordinates[j]
            except IndexError:
                loc_2 = ''
            print(
                f'{name_1} {i} {loc_1} and {name_2} {j} {loc_2} anticommuting'
            )
            operator_spec_1 = operator_spec(code, operators_1[i])
            operator_spec_2 = operator_spec(code, operators_2[j])
            overlap = [
                (op_1, op_2, site_1)
                for op_1, site_1 in operator_spec_1
                for op_2, site_2 in operator_spec_2
                if site_1 == site_2
            ]
            print('Overlap:', overlap)
            print(f'{name_1} {i}:', operator_spec_1)
            print(f'{name_2} {j}:', operator_spec_2)
            if i_print == max_print:
                n_remaining = len(non_commuting) - 1 - i_print
                if n_remaining > 0:
                    print(f'... {n_remaining} more pairs')
                break



import numpy as np

from flask import (
    Flask, send_from_directory, request, json, render_template
)
from panqec.codes import (
    Toric2DCode, RotatedPlanar2DCode, Planar2DCode,
    Toric3DCode, RotatedPlanar3DCode, RhombicToricCode, RhombicPlanarCode,
    Planar3DCode, RotatedToric3DCode, XCubeCode,
    HollowPlanar3DCode, HollowRhombicCode,
    Color3DCode, Color666PlanarCode, Color666ToricCode, Color488Code
)
from panqec.decoders import (
    MatchingDecoder, RotatedSweepMatchDecoder,
    SweepMatchDecoder, BeliefPropagationOSDDecoder,
    MemoryBeliefPropagationDecoder, XCubeMatchingDecoder
)
from panqec.error_models import PauliErrorModel

import webbrowser
import argparse

codes = {'Toric 2D': Toric2DCode,
         'Planar 2D': Planar2DCode,
         'Rotated Planar 2D': RotatedPlanar2DCode,
         '6.6.6 Color Code (toric)': Color666ToricCode,
         '6.6.6 Color Code (planar)': Color666PlanarCode,
         '4.8.8 Color Code': Color488Code,
         'Toric 3D': Toric3DCode,
         'Planar 3D': Planar3DCode,
         'Rotated Toric 3D': RotatedToric3DCode,
         'Rotated Planar 3D': RotatedPlanar3DCode,
         'Rhombic Toric 3D': RhombicToricCode,
         'Rhombic Planar 3D': RhombicPlanarCode,
         'XCube': XCubeCode,
         'Hollow Planar 3D': HollowPlanar3DCode,
         'Hollow Rhombic Code': HollowRhombicCode,
         '3D Color Code': Color3DCode}

noise_directions = {'Pure X': (1, 0, 0),
                    'Pure Y': (0, 1, 0),
                    'Pure Z': (0, 0, 1),
                    'Depolarizing': (1/3, 1/3, 1/3)}

decoders = {'BP-OSD': BeliefPropagationOSDDecoder,
            'MBP': MemoryBeliefPropagationDecoder,
            'SweepMatch': SweepMatchDecoder,
            'RotatedSweepMatch': RotatedSweepMatchDecoder,
            'Matching': MatchingDecoder,
            'XCube Matching': XCubeMatchingDecoder}


class GUI():
    app = None

    def __init__(self):
        self.app = Flask(__name__)

        self.codes = codes
        self.decoders = decoders
        self._code_names = {}

        self.add_all_routes()

    @property
    def code_names(self):
        self._code_names = {
            '2d': [code[0] for code in self.codes.items()
                   if code[1].dimension == 2],
            '3d': [code[0] for code in self.codes.items()
                   if code[1].dimension == 3]
        }

        return self._code_names

    @property
    def decoder_names(self):
        self._decoder_names = list(self.decoders.keys())

        return self._decoder_names

    def add_code(self, code_class, code_name):
        self.codes[code_name] = code_class

    def add_decoder(self, decoder_class, decoder_name):
        self.decoders[decoder_name] = decoder_class

    def _instantiate_code(self, data):
        Lx = data['Lx']
        Ly = data['Ly']
        if 'Lz' in data:
            Lz = data['Lz']
        code_name = data['code_name']
        deformation_name = data['code_deformation_name']

        if code_name in self.code_names['2d']:
            code = codes[code_name](Lx, Ly)
        elif code_name in self.code_names['3d']:
            code = codes[code_name](Lx, Ly, Lz)
        else:
            raise ValueError(f'Code {code_name} not recognized')

        if deformation_name != "None":
            code.deform(deformation_name)

        return code

    def run(self, *args, **kwargs):
        self.app.run(*args, **kwargs)

    def add_all_routes(self):
        self.app.add_url_rule(
            "/", "send_index",
            self.send_index
        )
        self.app.add_url_rule(
            "/main.css", "css",
            self.css
        )
        self.app.add_url_rule(
            "/favicon.ico", "favicon",
            self.favicon
        )
        self.app.add_url_rule(
            "/js/<path:path>", "send_js",
            self.send_js
        )
        self.app.add_url_rule(
            "/2d", "send_index_2d",
            self.send_index_2d
        )
        self.app.add_url_rule(
            "/3d", "send_index_3d",
            self.send_index_3d
        )
        self.app.add_url_rule(
            "/code-names", "code-names",
            self.send_code_names, methods=['POST']
        )
        self.app.add_url_rule(
            "/decoder-names", "decoder-names",
            self.send_decoder_names, methods=['POST']
        )
        self.app.add_url_rule(
            "/deformation-names", "deformation-names",
            self.send_deformation_names, methods=['POST']
        )
        self.app.add_url_rule(
            "/code-data", "code-data",
            self.send_code_data, methods=['POST']
        )
        self.app.add_url_rule(
            "/decode", "decode",
            self.send_correction, methods=['POST']
        )
        self.app.add_url_rule(
            "/new-errors", "new-errors",
            self.send_random_errors, methods=['POST']
        )

    def send_index_2d(self):
        return render_template('gui.html')

    def send_index_3d(self):
        return render_template('gui.html')

    def send_index(self):
        return render_template('index.html')

    def css(self):
        return send_from_directory('static/css', 'main.css')

    def favicon(self):
        return send_from_directory('static', 'favicon.ico')

    def send_js(self, path):
        return send_from_directory('js', path)

    def send_code_names(self):
        dim = request.json['dimension']
        code_names = self.code_names[f'{dim}d']

        return json.dumps(code_names)

    def send_decoder_names(self):
        code_name = request.json['code_name']
        code_id = codes[code_name].__name__

        decoder_names = []

        for decoder_name, decoder_class in decoders.items():
            if (decoder_class.allowed_codes is None
                    or code_id in decoder_class.allowed_codes):
                decoder_names.append(decoder_name)

        return json.dumps(decoder_names)

    def send_deformation_names(self):
        code_name = request.json['code_name']
        deformation_names = codes[code_name].deformation_names

        return json.dumps(deformation_names)

    def send_code_data(self):
        rotated_picture = request.json['rotated_picture']

        code = self._instantiate_code(request.json)

        qubits = [code.qubit_representation(location, rotated_picture)
                  for location in code.qubit_coordinates]
        stabilizers = [code.stabilizer_representation(location,
                                                      rotated_picture)
                       for location in code.stabilizer_coordinates]

        logical_z = code.logicals_z
        logical_x = code.logicals_x

        return json.dumps({'H': code.stabilizer_matrix.toarray().tolist(),
                           'qubits': qubits,
                           'stabilizers': stabilizers,
                           'logical_z': logical_z.tolist(),
                           'logical_x': logical_x.tolist()
                           })

    def send_correction(self):
        content = request.json
        syndrome = np.array(content['syndrome'])
        p = content['p']
        noise_deformation = content['noise_deformation_name']
        max_bp_iter = content['max_bp_iter']
        alpha = content['alpha']
        beta = content['beta']
        decoder_name = content['decoder']
        error_model_name = content['error_model']

        if noise_deformation == 'None':
            noise_deformation = None

        code = self._instantiate_code(content)

        rx, ry, rz = noise_directions[error_model_name]
        error_model = PauliErrorModel(rx, ry, rz, noise_deformation)

        kwargs = {}
        if decoder_name in ['BP-OSD', 'MBP']:
            kwargs['max_bp_iter'] = max_bp_iter
        if decoder_name == 'BP-OSD':
            kwargs['osd_order'] = 0
        if decoder_name == 'MBP':
            kwargs['alpha'] = alpha
            kwargs['beta'] = beta

        decoder = self.decoders[decoder_name](code, error_model, p,
                                              **kwargs)

        correction = decoder.decode(syndrome)

        correction_x = correction[:code.n]
        correction_z = correction[code.n:]

        return json.dumps({'x': correction_x.tolist(),
                           'z': correction_z.tolist()})

    def send_random_errors(self):
        content = request.json
        p = content['p']
        noise_deformation = content['noise_deformation_name']
        error_model_name = content['error_model']

        if noise_deformation == 'None':
            noise_deformation = None

        code = self._instantiate_code(content)

        rx, ry, rz = noise_directions[error_model_name]
        error_model = PauliErrorModel(rx, ry, rz, noise_deformation)

        errors = error_model.generate(code, p)

        bsf_to_str_map = {(0, 0): 'I', (1, 0): 'X', (0, 1): 'Z',
                          (1, 1): 'Y'}
        error_spec = [
            (
                bsf_to_str_map[
                    (errors[i_qubit], errors[i_qubit + code.n])
                ],
                [
                    coords for index, coords in enumerate(
                        code.qubit_coordinates
                    )
                    if index == i_qubit
                ][0]
            )
            for i_qubit in range(code.n)
        ]
        error_spec = [spec for spec in error_spec if spec[0] != 'I']
        return json.dumps(errors.tolist())


def open_browser(port):
    webbrowser.open_new_tab(f'http://127.0.0.1:{port}/')


def run_gui():
    parser = argparse.ArgumentParser(description='Run GUI server')
    parser.add_argument('--port', '-p', type=int, default=5000,
                        help='Port where to run the server')
    args = parser.parse_args()
    port = args.port

    gui = GUI()
    # gui.add_all_endpoints()
    gui.run(port=port)


if __name__ == '__main__':
    run_gui()



"""
Bit array or vector representations of Paulis for 3Di codes.

Although qecsim already has such an implementation, some of these extra
routines are useful specifically for dealing with the 3D code.

:Author:
    Eric Huang
"""
from typing import Union, List
import numpy as np
from . import bsparse
from scipy.sparse import csr_matrix


def bs_prod(a, b) -> np.ndarray:
    """Array of 0 for commutes and 1 for anticommutes bvectors."""

    # If lists, convert to numpy
    if isinstance(a, list):
        a = np.array(a, dtype='uint8')
    if isinstance(b, list):
        b = np.array(b, dtype='uint8')

    # Determine the output shape.
    # In particular, flatten array where needed.
    output_shape = None
    if len(a.shape) == 2 and len(b.shape) == 1:
        output_shape = a.shape[0]
    elif len(a.shape) == 1 and len(b.shape) == 2:
        output_shape = b.shape[0]

    # If only singles, then convert to 2D array.
    if len(a.shape) == 1:
        a = np.reshape(a, (1, a.shape[0]))
    if len(b.shape) == 1:
        b = np.reshape(b, (1, b.shape[0]))

    # Check the shapes are correct.
    if a.shape[1] % 2 != 0:
        raise ValueError(
            f'Length {a.shape[1]} binary vector not of even length.'
        )
    if b.shape[1] % 2 != 0:
        raise ValueError(
            f'Length {b.shape[1]} binary vector not of even length.'
        )
    if b.shape[1] != a.shape[1]:
        raise ValueError(
            f'Length {a.shape[1]} bvector cannot be '
            f'composed with length {b.shape[1]}'
        )

    if bsparse.is_sparse(a) or bsparse.is_sparse(b):
        commutes = _bs_prod_sparse(a, b)
    else:
        # Number of qubits.
        n = int(a.shape[1]/2)

        # Commute commutator by binary symplectic form.
        a_X = a[:, :n]
        a_Z = a[:, n:]
        b_X = b[:, :n]
        b_Z = b[:, n:]

        commutes = (a_X.dot(b_Z.T) + a_Z.dot(b_X.T)) % 2

    if output_shape is not None:
        commutes = commutes.reshape(output_shape)

    return commutes


def _bs_prod_sparse(a, b):
    """Array of 0 for commutes and 1 for anticommutes bvectors."""

    # Commute commutator by binary symplectic form.
    n = int(a.shape[1]/2)

    if not bsparse.is_sparse(a):
        a = bsparse.from_array(a)
    if not bsparse.is_sparse(b):
        b = bsparse.from_array(b)

    a_X = a[:, :n]
    a_Z = a[:, n:]
    b_X = b[:, :n]
    b_Z = b[:, n:]

    commutes = (a_X.dot(b_Z.T) + a_Z.dot(b_X.T))
    commutes.data %= 2

    if commutes.shape[0] == 1:
        return commutes.toarray()[0, :]
    elif commutes.shape[1] == 1:
        return commutes.toarray()[:, 0]
    else:
        return commutes.toarray()


def pauli_to_bsf(error_pauli):
    ps = np.array(list(error_pauli))
    xs = (ps == 'X') + (ps == 'Y')
    zs = (ps == 'Z') + (ps == 'Y')

    error = np.hstack((xs, zs)).astype('uint8')

    return error


def pauli_string_to_bvector(pauli_string: str) -> np.ndarray:
    X_block = []
    Z_block = []
    for character in pauli_string:
        if character == 'I':
            X_block.append(0)
            Z_block.append(0)
        elif character == 'X':
            X_block.append(1)
            Z_block.append(0)
        elif character == 'Y':
            X_block.append(1)
            Z_block.append(1)
        elif character == 'Z':
            X_block.append(0)
            Z_block.append(1)
    bvector = np.concatenate([X_block, Z_block]).astype(np.uint)
    return bvector


def bvector_to_pauli_string(bvector: np.ndarray) -> str:
    n = int(bvector.shape[0]/2)
    pauli_string = ''
    for i in range(n):
        pauli_string += {
            (0, 0): 'I',
            (1, 0): 'X',
            (1, 1): 'Y',
            (0, 1): 'Z'
        }[(bvector[i], bvector[i + n])]
    return pauli_string


def get_effective_error(
    total_error,
    logicals_x,
    logicals_z,
) -> np.ndarray:
    """Effective Pauli error on logical qubits after decoding."""

    if logicals_x.shape != logicals_z.shape:
        raise ValueError('Logical Xs and Zs must be of same shape.')

    # Number of pairs of logical operators.
    if len(logicals_x.shape) == 1:
        n_logical = 1
    else:
        n_logical = int(logicals_x.shape[0])

    # Get the number of total errors given.
    num_total_errors = 1
    if len(total_error.shape) > 1:
        num_total_errors = total_error.shape[0]

    # The shape of the array to be returned.
    final_shape: tuple = (num_total_errors, 2*n_logical)
    if num_total_errors == 1:
        final_shape = (2*n_logical, )

    effective_Z = bs_prod(logicals_x, total_error)
    effective_X = bs_prod(logicals_z, total_error)

    if num_total_errors == 1:
        effective = np.concatenate([effective_X, effective_Z])
    elif n_logical == 1:
        effective = np.array([effective_X, effective_Z]).T
    else:
        effective = np.array([
            np.concatenate([effective_X[:, i], effective_Z[:, i]])
            for i in range(num_total_errors)
        ])

    # Flatten the array if only one total error is given.
    effective = effective.reshape(final_shape)
    return effective


def bvector_to_int(bvector: np.ndarray) -> int:
    """Convert bvector to integer for effecient storage."""
    return int(''.join(map(str, bvector)), 2)


def int_to_bvector(int_rep: int, n: int) -> np.ndarray:
    """Convert integer representation to n-qubit Pauli bvector."""
    binary_string = ('{:0%db}' % (2*n)).format(int_rep)
    bvector = np.array(tuple(binary_string), dtype=np.uint)
    return bvector


def bvectors_to_ints(bvector_list: list) -> list:
    """List of bvectors to integers for efficient storage."""
    return list(map(
        bvector_to_int,
        bvector_list
    ))


def ints_to_bvectors(int_list: list, n: int) -> list:
    """Convert list of integers back to bvectors."""
    bvectors = []
    for int_rep in int_list:
        bvectors.append(int_to_bvector(int_rep, n))
    return bvectors


def gf2_rank(rows):
    """Find rank of a matrix over GF2 given as list of binary ints.

    From https://stackoverflow.com/questions/56856378
    """
    rank = 0
    while rows:
        pivot_row = rows.pop()
        if pivot_row:
            rank += 1
            lsb = pivot_row & -pivot_row
            for index, row in enumerate(rows):
                if row & lsb:
                    rows[index] = row ^ pivot_row
    return rank


def brank(matrix):
    """Rank of a binary matrix."""

    matrix = bsparse.to_array(matrix)

    # Convert to list of binary numbers.
    rows = [int(''.join(map(str, row)), 2) for row in matrix.astype(int)]
    return gf2_rank(rows)


def apply_deformation(
    deformation_indices: Union[List[bool], np.ndarray], bsf: np.ndarray
) -> np.ndarray:
    """Return Hadamard-deformed bsf at given indices."""
    n = len(deformation_indices)
    deformed = np.zeros_like(bsf)
    if len(bsf.shape) == 1:
        if bsf.shape[0] != 2*n:
            raise ValueError(
                f'Deformation index length {n} does not match '
                f'bsf shape {bsf.shape}, which should be {(2*n,)}'
            )
        for i, deform in enumerate(deformation_indices):
            if deform:
                deformed[i] = bsf[i + n]
                deformed[i + n] = bsf[i]
            else:
                deformed[i] = bsf[i]
                deformed[i + n] = bsf[i + n]
    else:
        if bsf.shape[1] != 2*n:
            raise ValueError(
                f'Deformation index length {n} does not match '
                f'bsf shape {bsf.shape}, which should be '
                f'{(bsf.shape[0], 2*n)}.'
            )
        for i, deform in enumerate(deformation_indices):
            if deform:
                deformed[:, i] = bsf[:, i + n]
                deformed[:, i + n] = bsf[:, i]
            else:
                deformed[:, i] = bsf[:, i]
                deformed[:, i + n] = bsf[:, i + n]
    return deformed


def bsf_wt(bsf):
    """
    Return weight of given binary symplectic form.
    :param bsf: Binary symplectic vector or matrix.
    :type bsf: numpy.array (1d or 2d) or csr_matrix
    :return: Weight
    :rtype: int
    """
    if isinstance(bsf, np.ndarray):
        assert np.array_equal(bsf % 2, bsf), \
                'BSF {} is not in binary form'.format(bsf)
        return np.count_nonzero(sum(np.hsplit(bsf, 2)))

    elif isinstance(bsf, csr_matrix):
        assert np.all(bsf.data == 1), \
                'BSF {} is not in binary form'.format(bsf)

        n = bsf.shape[1] // 2
        x_indices = bsf.indices[bsf.indices < n]
        z_indices = bsf.indices[bsf.indices >= n] - n

        return len(np.union1d(x_indices, z_indices))
    else:
        raise TypeError(
            f"bsf matrix should be a numpy array or "
            f"csr_matrix, not {type(bsf)}"
        )


def bsf_to_pauli(bsf):
    """
    Convert the given binary symplectic form to Pauli operator(s).
    (1 0 0 0 1 | 0 0 1 0 1) -> XIZIY
    Assumptions:
    * bsf is a numpy.array (1d or 2d) in binary symplectic form.
    :param bsf: Binary symplectic vector or matrix.
    :type bsf: numpy.array (1d or 2d)
    :return: Pauli operators.
    :rtype: str or list of str
    """

    if isinstance(bsf, np.ndarray):
        assert np.array_equal(bsf % 2, bsf), \
                'BSF {} is not in binary form'.format(bsf)

        def _to_pauli(b, t=str.maketrans('0123', 'IXZY')):  # noqa: B008,E501 (deliberately reuse t)
            xs, zs = np.hsplit(b, 2)
            ps = (xs + zs * 2).astype(str)  # 0=I, 1=X, 2=Z, 3=Y
            return ''.join(ps).translate(t)

        if bsf.ndim == 1:
            return _to_pauli(bsf)
        else:
            return [_to_pauli(b) for b in bsf]
    else:
        assert np.all(bsf.data == 1), \
                'BSF {} is not in binary form'.format(bsf)

        def _to_pauli(b):  # type:ignore
            n = bsf.shape[1] // 2
            pauli_string = ['I' for _ in range(n)]
            for i in b.indices:
                if i < n:
                    pauli_string[i] = 'X'
                elif i >= n:
                    if pauli_string[i - n] == 'X':
                        pauli_string[i - n] = 'Y'
                    else:
                        pauli_string[i - n] = 'Z'

            return ''.join(pauli_string)

        return [_to_pauli(b) for b in bsf]



import numpy as np
import pytest
from panqec.bpauli import bsf_wt
from panqec.codes import Toric3DCode
import panqec.bsparse as bsparse
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestToric3DCode(StabilizerCodeTest):

    @pytest.fixture(params=[(2, 2, 2), (3, 3, 3), (2, 3, 4)])
    def code(self, request):
        return Toric3DCode(*request.param)

    def test_get_vertex_stabilizers(self, code):
        n = code.n

        vertex_index = tuple(
            index
            for index, location in enumerate(code.stabilizer_coordinates)
            if code.stabilizer_type(location) == 'vertex'
        )
        stabilizers = code.stabilizer_matrix[vertex_index, :]

        # There should be least some vertex stabilizers.
        assert stabilizers.shape[0] > 0

        # For sparse matrices, shape[1] matters
        assert len(stabilizers.shape) == 1 or stabilizers.shape[1] > 0

        assert stabilizers.dtype == 'uint8'

        # All Z stabilizers should be weight 6.
        assert all(
            bsf_wt(stabilizer) == 6 for stabilizer in stabilizers
        )

        # Number of stabilizer generators should be number of vertices.
        assert stabilizers.shape[0] == np.product(code.size)

        # Each bsf length should be 2*n.
        assert stabilizers.shape[1] == 2*n

        # There should be no X or Y operators.
        if isinstance(stabilizers, np.ndarray):
            assert np.all(stabilizers[:, :n] == 0)
        else:
            assert bsparse.equal(stabilizers[:, :n], 0)

        assert all(
            'X' not in code.from_bsf(stabilizer)
            and 'Y' not in code.from_bsf(stabilizer)
            for stabilizer in stabilizers
        )

        # Each qubit should be in the support of exactly 2 stabilizers.
        assert np.all(stabilizers.sum(axis=0)[n:] == 2)

    def test_general_properties(self, code):
        n, k, d = code.n, code.k, code.d

        # The number of qubits should be the number of edges 3*L_x*L_y*L_z.
        assert n == 3*np.product(code.size)
        assert k == 3
        assert d == min(code.size)

    def test_get_face_stabilizers(self, code):
        n = code.n
        face_index = tuple(
            index
            for index, location in enumerate(code.stabilizer_coordinates)
            if code.stabilizer_type(location) == 'face'
        )
        stabilizers = code.stabilizer_matrix[face_index, :]

        # Weight of every stabilizer should be 4.
        assert np.all(stabilizers.sum(axis=1) == 4)
        assert stabilizers.dtype == 'uint8'

        # Number of stabilizer generators should be number of edges.
        assert stabilizers.shape[0] == 3*np.product(code.size)

        # The number of qubits should be the number of edges 3L^3.
        assert stabilizers.shape[1] == 2*n

        # There should be no Z or Y operators.
        if isinstance(stabilizers, np.ndarray):
            assert np.all(stabilizers[:, n:] == 0)
        else:
            assert bsparse.equal(stabilizers[:, n:], 0)

        # Each qubit should be in the support of exactly 4 stabilizers.
        if isinstance(stabilizers, np.ndarray):
            assert np.all(stabilizers.sum(axis=0)[:n] == 4)
        else:
            assert np.all(np.array(stabilizers.sum(axis=0)[0, :n]) == 4)

    def test_get_all_stabilizers(self, code):
        n = code.n
        n_vertex = len(code.type_index('vertex'))
        stabilizers = code.stabilizer_matrix

        # Total number of stabilizers.
        assert stabilizers.shape[0] == 4*np.product(code.size)

        # Z block of X stabilizers should be all 0.
        assert np.all(bsparse.to_array(stabilizers[:n_vertex, :n]) == 0)

        # X block of Z stabilizers should be all 0.
        assert np.all(
            bsparse.to_array((stabilizers[n_vertex:, n:])) == 0
        )

    def test_get_Z_logicals(self, code):
        n = code.n
        logicals = code.logicals_z
        assert logicals.shape[0] == 3
        assert logicals.shape[1] == 2*n

    def test_get_X_logicals(self, code):
        n = code.n
        logicals = code.logicals_x
        assert logicals.shape[0] == 3
        assert logicals.shape[1] == 2*n



"""
Settings from environmental variables and config files.

:Author:
    Eric Huang
"""
import os
from typing import Dict, Any
from dotenv import load_dotenv
from .codes import (
    Toric2DCode, Planar2DCode, RotatedPlanar2DCode,
    Toric3DCode, RotatedPlanar3DCode, RotatedToric3DCode,
    Color666PlanarCode, Color666ToricCode, Color488Code,
    Color3DCode, RhombicToricCode, RhombicPlanarCode,
    XCubeCode, HollowPlanar3DCode, HollowRhombicCode
)
from .decoders import (
    SweepMatchDecoder, XCubeMatchingDecoder,
    RotatedSweepMatchDecoder
)
from .decoders import BeliefPropagationOSDDecoder
from .decoders import MemoryBeliefPropagationDecoder
from .decoders.matching._matching_decoder import MatchingDecoder
from .error_models import PauliErrorModel

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

# Load the .env file into environmental variables.
if os.getenv('PANQEC_DIR') is None:
    load_dotenv()

PANQEC_DARK_THEME = False
if os.getenv('PANQEC_DARK_THEME'):
    PANQEC_DARK_THEME = bool(os.getenv('PANQEC_DARK_THEME'))

# Fallback is to use temp dir inside repo if PANQEC_DIR is not available.
PANQEC_DIR = os.path.join(
    os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
    'temp'
)

# Load the output directory from environmental variables.
if os.getenv('PANQEC_DIR') is not None:
    PANQEC_DIR = os.path.abspath(str(os.getenv('PANQEC_DIR')))
    if not os.path.isdir(PANQEC_DIR):
        raise FileNotFoundError(
            f'PANQEC_DIR={PANQEC_DIR} is not a valid directory. '
            'Check .env configuration.'
        )
else:
    os.makedirs(PANQEC_DIR, exist_ok=True)

# Register your models here.
CODES = {
    'Toric2DCode': Toric2DCode,
    'Planar2DCode': Planar2DCode,
    'RotatedPlanar2DCode': RotatedPlanar2DCode,
    'Color666PlanarCode': Color666PlanarCode,
    'Color666ToricCode': Color666ToricCode,
    'Color488Code': Color488Code,
    'Color3DCode': Color3DCode,
    'Toric3DCode': Toric3DCode,
    'Planar3DCode': RotatedPlanar3DCode,
    'RotatedPlanar3DCode': RotatedPlanar3DCode,
    'RotatedToric3DCode': RotatedToric3DCode,
    'RhombicToricCode': RhombicToricCode,
    'RhombicPlanarCode': RhombicPlanarCode,
    'XCubeCode': XCubeCode,
    'HollowPlanar3DCode': HollowPlanar3DCode,
    'HollowRhombicCode': HollowRhombicCode
}
ERROR_MODELS = {
    'PauliErrorModel': PauliErrorModel
}
DECODERS: Dict[str, Any] = {
    'MatchingDecoder': MatchingDecoder,
    'SweepMatchDecoder': SweepMatchDecoder,
    'RotatedSweepMatchDecoder': RotatedSweepMatchDecoder,
    'BeliefPropagationOSDDecoder': BeliefPropagationOSDDecoder,
    'MemoryBeliefPropagationDecoder': MemoryBeliefPropagationDecoder,
    'XCubeMatchingDecoder': XCubeMatchingDecoder
}

# Slurm automation config.
SLURM_DIR = os.path.join(os.path.dirname(BASE_DIR), 'slurm')
if os.getenv('SLURM_DIR') is not None:
    SLURM_DIR = os.path.abspath(str(os.getenv('SLURM_DIR')))

SBATCH_TEMPLATE = os.path.join(
    os.path.dirname(BASE_DIR), 'scripts', 'template.sbatch'
)
AD_TEMPLATE = os.path.join(
    os.path.dirname(BASE_DIR), 'scripts', 'ad.sbatch'
)

# Slurm username for reporting status.
SLURM_USERNAME = None
if os.getenv('USER') is not None:
    SLURM_USERNAME = os.getenv('USER')
elif os.getenv('USERNAME') is not None:
    SLURM_USERNAME = os.getenv('USERNAME')


def register_code(code_class):
    label = code_class.__class__.__name__
    CODES[label] = code_class


def register_error_model(error_model_class):
    label = error_model_class.__class__.__name__
    ERROR_MODELS[label] = error_model_class


def register_decoder(decoder_class):
    label = decoder_class.__class__.__name__
    DECODERS[label] = decoder_class


# Shortened names for analysis and plot labelling purposes.
SHORT_NAMES = {
    'Pauli': 'CSS',
    'Deformed XZZX Pauli': 'Clifford-deformed',
    'Deformed Rhombic Pauli': 'Clifford-deformed',
    'BP-OSD decoder': 'BP-OSD',
    'Toric 3D Sweep Pymatching Decoder': 'sweep-matching',
    'Deformed Toric 3D Sweep Pymatching Decoder': 'sweep-matching',
}
LONG_NAMES = {
    'Rhombic': 'surface code on checkerboard lattice',
    'Toric': 'surface code on cubic lattice',
    'XCube': 'X-cube model',
}



import pytest
from typing import List
from abc import ABCMeta, abstractmethod
import numpy as np
from panqec.codes import StabilizerCode
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel, PauliErrorModel


class DecoderTest(metaclass=ABCMeta):

    @pytest.fixture
    @abstractmethod
    def code(self) -> StabilizerCode:
        pass

    @pytest.fixture
    def error_model(self) -> BaseErrorModel:
        return PauliErrorModel(1/3, 1/3, 1/3)

    @pytest.fixture
    @abstractmethod
    def decoder(self, code, error_model) -> BaseDecoder:
        pass

    @pytest.fixture
    def allowed_paulis(self) -> List[str]:
        return ['X', 'Y', 'Z']

    def test_decode_trivial_syndrome(self, code, decoder):
        syndrome = np.zeros(
            shape=code.stabilizer_matrix.shape[0], dtype=np.uint
        )
        correction = decoder.decode(syndrome)
        assert correction.shape[0] == 2*code.n
        assert np.all(code.measure_syndrome(correction) == 0)
        assert np.all(correction == 0)
        assert code.is_success(correction)
        assert code.in_codespace(correction)

    @pytest.mark.slow
    def test_decode_single_qubit_error(self, code, decoder, allowed_paulis):
        for pauli in allowed_paulis:
            for i in range(code.n):
                error = np.zeros(2*code.n, dtype='uint8')

                if pauli in ['X', 'Y']:
                    error[i] = 1
                if pauli in ['Z', 'Y']:
                    error[code.n + i] = 1

                syndrome = code.measure_syndrome(error)

                correction = decoder.decode(syndrome)
                total_error = (correction + error) % 2

                assert code.in_codespace(total_error), (
                    "The decoding result is outside of the codespace"
                    f"for single-qubit {pauli} error at qubit {i}"
                )

                assert code.is_success(total_error), (
                    "The decoding result is has created a logical"
                    f"for single-qubit {pauli} error at qubit {i}"
                )



from ._toric_3d_code import Toric3DCode  # noqa
from ._planar_3d_code import Planar3DCode  # noqa
from ._rotated_planar_3d_code import RotatedPlanar3DCode  # noqa
from ._hollow_planar_3d_code import HollowPlanar3DCode  # noqa
from ._rhombic_toric_code import RhombicToricCode  # noqa
from ._rhombic_planar_code import RhombicPlanarCode  # noqa
from ._hollow_rhombic_code import HollowRhombicCode
from ._rotated_toric_3d_code import RotatedToric3DCode  # noqa

__all__ = [
    'Toric3DCode',
    'Planar3DCode',
    'RotatedPlanar3DCode',
    'HollowPlanar3DCode',
    'RhombicToricCode',
    'RhombicPlanarCode',
    'HollowRhombicCode',
    'RotatedToric3DCode',
]



from panqec.io import serialize_results, dump_results
import numpy as np
import datetime
import os
import pytest


@pytest.fixture
def serialized_results():
    i_trial = 2
    n_trials = 10

    L_list = np.array([1, 2])
    p_list = np.array([0.1, 0.2])
    L_repeats = np.array([1, 1])

    start_time = datetime.datetime(2021, 4, 6, 20, 41, 26, 887933)
    time_elapsed = datetime.timedelta(seconds=10699, microseconds=324064)

    time_remaining = datetime.timedelta(0)
    eta = datetime.datetime(2021, 4, 6, 20, 41, 26, 887933)

    np.random.seed(0)
    p_est = np.random.rand(L_list.shape[0], p_list.shape[0])
    p_se = np.random.rand(L_list.shape[0], p_list.shape[0])

    n_fail = np.random.randint(0, n_trials + 1, size=p_est.shape)
    n_try = np.ones(p_est.shape, dtype=int)*10

    effective_errors = [
        [
            [
                np.random.randint(0, 2, size=3*L_list[i_L]**3)
                for i_run in range(len(p_list))
            ]
            for i_p in range(len(p_list))
        ]
        for i_L in range(len(L_list))
    ]

    results_dict = serialize_results(
        i_trial, n_trials, L_list, p_list, L_repeats, start_time, time_elapsed,
        time_remaining, eta, p_est, p_se, n_fail, n_try, effective_errors
    )

    assert 'parameters' in results_dict.keys()
    assert 'time' in results_dict.keys()
    assert 'statistics' in results_dict.keys()
    assert 'results' in results_dict.keys()
    return results_dict


def test_dump_results(serialized_results, tmpdir):
    results_dict = serialized_results
    p = tmpdir.mkdir('sub')
    export_json = os.path.join(p, 'example.json')
    assert not os.path.exists(export_json)

    dump_results(export_json, results_dict)
    assert os.path.exists(export_json)



import pytest
import numpy as np
from panqec.codes import RotatedPlanar3DCode
from tests.codes.stabilizer_code_test import (
    StabilizerCodeTest, StabilizerCodeTestWithCoordinates
)


class RotatedPlanar3DCodeTest(StabilizerCodeTestWithCoordinates):
    code_class = RotatedPlanar3DCode


class TestRotatedPlanar3DCode(StabilizerCodeTest):

    L_x = 4
    L_y = 4
    L_z = 4

    @pytest.fixture
    def code(self):
        """Example code with non-uniform dimensions."""
        new_code = RotatedPlanar3DCode(self.L_x, self.L_y, self.L_z)
        return new_code

    def test_vertex_index_corner_region(self, code):

        vertex_index = code.type_index('vertex')
        # First layer corner near origin
        assert (0, 0, 0) not in vertex_index
        assert (2, 0, 1) in vertex_index
        assert (2, 4, 1) in vertex_index
        assert (0, 2, 1) not in vertex_index
        assert (4, 2, 1) in vertex_index
        assert (6, 0, 1) in vertex_index
        assert (2, 8, 1) in vertex_index
        assert (4, 6, 1) in vertex_index

    def test_vertex_index_complies_with_rules(self, code):
        for x, y, z in code.type_index('vertex'):
            assert z % 2 == 1
            if x % 4 == 2:
                assert y % 4 == 0
            else:
                assert x % 4 == 0
                assert y % 4 == 2

    def test_vertex_index_boundary_conditions(self, code):
        for x, y, z in code.type_index('vertex'):
            assert x != 0
            if y == 0:
                assert x % 4 == 2
            if self.L_x % 2 == 0:
                if x == self.L_x*2 + 2:
                    assert y % 4 == 0

    def test_face_index_complies_with_rules(self, code):
        for x, y, z in code.type_index('face'):
            if z % 2 == 1:
                if x % 4 == 0:
                    assert y % 4 == 0
                else:
                    assert x % 4 == 2
                    assert y % 4 == 2
            else:
                assert x % 2 == 1
                assert y % 2 == 1

    def test_qubit_coordinates_complies_with_rules(self, code):
        for x, y, z in code.qubit_coordinates:
            if z % 2 == 1:
                assert x % 2 == 1
                assert y % 2 == 1
            else:
                if x % 4 == 0:
                    assert y % 4 == 2
                else:
                    assert x % 4 == 2
                    assert y % 4 == 0

    def test_each_qubit_contained_in_1_or_2_check_operators(self, code):
        H = code.Hz
        assert np.all(H.sum(axis=0) > 0)
        assert np.all(H.sum(axis=0) <= 2)


class TestRotatedPlanar3DCode3x3x3(RotatedPlanar3DCodeTest):
    size = (3, 3, 3)
    expected_plane_edges_xy = [
        (1, 1), (1, 3), (1, 5),
        (3, 1), (3, 3), (3, 5),
        (5, 1), (5, 3), (5, 5),
    ]
    expected_vertical_faces_xy = expected_plane_edges_xy
    expected_plane_faces_xy = [
        (0, 4), (2, 2), (4, 4), (6, 2)
    ]
    expected_plane_vertices_xy = [
        (2, 0), (2, 4),
        (4, 2), (4, 6),
    ]
    expected_plane_z = [1, 3, 5]
    expected_vertical_z = [2, 4]


class TestRotatedPlanar3DCode4x4x4(RotatedPlanar3DCodeTest):
    size = (4, 4, 4)
    expected_plane_edges_xy = [
        (1, 1), (1, 3), (1, 5), (1, 7),
        (3, 1), (3, 3), (3, 5), (3, 7),
        (5, 1), (5, 3), (5, 5), (5, 7),
        (7, 1), (7, 3), (7, 5), (7, 7),
    ]
    expected_vertical_faces_xy = expected_plane_edges_xy
    expected_plane_faces_xy = [
        (0, 4),
        (2, 2), (2, 6),
        (4, 4),
        (6, 2), (6, 6),
        (8, 4),
    ]
    expected_plane_vertices_xy = [
        (2, 0), (2, 4), (2, 8),
        (4, 2), (4, 6),
        (6, 0), (6, 4), (6, 8)
    ]
    expected_plane_z = [1, 3, 5, 7]
    expected_vertical_z = [2, 4, 6]



import numpy as np
import pytest
from panqec.bpauli import bsf_to_pauli, bsf_wt
from panqec.error_models import PauliErrorModel
from panqec.codes import Toric3DCode
from panqec.bsparse import to_array
from panqec.utils import get_direction_from_bias_ratio


class TestPauliNoise:

    @pytest.fixture(autouse=True)
    def seed_random(self):
        np.random.seed(0)

    @pytest.fixture
    def code(self):
        return Toric3DCode(3, 4, 5)

    @pytest.fixture
    def error_model(self):
        return PauliErrorModel(0.2, 0.3, 0.5)

    def test_label(self, error_model):
        assert error_model.label == 'Pauli X0.2000Y0.3000Z0.5000'

    def test_generate(self, code, error_model):
        error_rate = 0.1
        error = to_array(
            error_model.generate(code, error_rate, rng=np.random)
        )
        assert np.any(error != 0), 'Error should be non-trivial'
        assert error.shape == (2*code.n, ), 'Shape incorrect'

    def test_probability_zero(self, code, error_model):
        error_rate = 0
        error = to_array(
            error_model.generate(code, error_rate, rng=np.random)
        )
        assert np.all(error == 0), 'Should have no error'

    def test_probability_one(self, code, error_model):
        error_rate = 1
        error = error_model.generate(code, error_rate, rng=np.random)

        # Error everywhere so weight is number of qubits.
        assert bsf_wt(error) == code.n, 'Should be error everywhere'

    def test_generate_all_X_errors(self, code):
        error_rate = 1
        direction = (1, 0, 0)
        error_model = PauliErrorModel(*direction)
        error = error_model.generate(code, error_rate, rng=np.random)
        assert bsf_to_pauli(error) == 'X'*code.n, (
            'Should be X error everywhere'
        )

    def test_generate_all_Y_errors(self, code):
        error_rate = 1
        direction = (0, 1, 0)
        error_model = PauliErrorModel(*direction)
        error = error_model.generate(code, error_rate, rng=np.random)
        assert bsf_to_pauli(error) == 'Y'*code.n, (
            'Should be Y error everywhere'
        )

    def test_generate_all_Z_errors(self, code):
        error_rate = 1
        direction = (0, 0, 1)
        error_model = PauliErrorModel(*direction)
        error = error_model.generate(code, error_rate, rng=np.random)
        assert bsf_to_pauli(error) == 'Z'*code.n, (
            'Should be Z error everywhere'
        )

    def test_raise_error_if_direction_does_not_sum_to_1(self):
        with pytest.raises(ValueError):
            PauliErrorModel(0, 0, 0)


class TestGeneratePauliNoise:

    def generate_pauli_noise(self, p_X, p_Y, p_Z, L):
        code = Toric3DCode(L, L, L)
        error_rate = p_X + p_Y + p_Z
        if error_rate != 0:
            r_x, r_y, r_z = p_X/error_rate, p_Y/error_rate, p_Z/error_rate
        else:
            r_x, r_y, r_z = 1/3, 1/3, 1/3
        error_model = PauliErrorModel(r_x, r_y, r_z)
        noise = error_model.generate(code, error_rate)
        return noise

    @pytest.fixture(autouse=True)
    def seed_random(self):
        np.random.seed(0)

    def test_generate_pauli_noise(self):
        p_X = 0.1
        p_Y = 0.2
        p_Z = 0.5
        L = 10
        noise = self.generate_pauli_noise(p_X, p_Y, p_Z, L)
        assert list(noise.shape) == [2*3*L**3]
        assert np.issubdtype(noise.dtype, np.unsignedinteger)

    def test_no_errors_if_p_zero(self):
        L = 10
        noise = self.generate_pauli_noise(0, 0, 0, L)
        assert np.all(noise == 0)

    def test_all_X_if_p_X_one(self):
        L = 10
        noise = self.generate_pauli_noise(1, 0, 0, L)
        assert np.all(noise[:3*L**3] == 1)
        assert np.all(noise[3*L**3:] == 0)

    def test_all_Y_if_p_Y_one(self):
        L = 10
        noise = self.generate_pauli_noise(0, 1, 0, L)
        assert np.all(noise[:3*L**3] == 1)
        assert np.all(noise[3*L**3:] == 1)

    def test_all_Z_if_p_Z_one(self):
        L = 10
        noise = self.generate_pauli_noise(0, 0, 1, L)
        assert np.all(noise[:3*L**3] == 0)
        assert np.all(noise[3*L**3:] == 1)

    def test_only_X_if_p_X_only(self):
        L = 10
        noise = self.generate_pauli_noise(0.5, 0, 0, L)
        assert np.any(noise[:3*L**3] == 1)
        assert np.all(noise[3*L**3:] == 0)

    def test_only_Y_if_p_Y_only(self):
        L = 10
        noise = self.generate_pauli_noise(0, 0.5, 0, L)
        assert np.all(noise[:3*L**3] == noise[3*L**3:])

    def test_only_Z_if_p_Z_only(self):
        L = 10
        noise = self.generate_pauli_noise(0, 0, 0.5, L)
        assert np.all(noise[:3*L**3] == 0)
        assert np.any(noise[3*L**3:] == 1)


class TestDeformedMatchingWeights:

    def get_weights(self, p_X, p_Y, p_Z, L):
        code = Toric3DCode(L, L, L)
        error_rate = p_X + p_Y + p_Z
        if error_rate != 0:
            r_x, r_y, r_z = p_X/error_rate, p_Y/error_rate, p_Z/error_rate
        else:
            r_x, r_y, r_z = 1/3, 1/3, 1/3
        error_model = PauliErrorModel(
            r_x, r_y, r_z,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        weights_x, weight_z = error_model.get_weights(code, error_rate)
        return weights_x, weight_z

    def test_if_equal_rates_then_equal_weights(self):
        L = 10
        p_X, p_Y, p_Z = 0.1, 0.1, 0.1
        wx, wz = self.get_weights(p_X, p_Y, p_Z, L)
        assert np.all(wx != 0)
        assert np.all(wx == wx[0])

        assert np.all(wz != 0)
        assert np.all(wz == wz[0])

    def test_zero_error_rate_no_nan(self):
        L = 10
        p_X, p_Y, p_Z = 0, 0, 0
        wx, wz = self.get_weights(p_X, p_Y, p_Z, L)
        assert np.all(wx != 0)
        assert np.all(~np.isnan(wx))

        assert np.all(wz != 0)
        assert np.all(~np.isnan(wz))

    def test_one_error_rate_no_nan(self):
        L = 10
        p_X, p_Y, p_Z = 0.45, 0, 0
        wx, wz = self.get_weights(p_X, p_Y, p_Z, L)
        assert np.all(wx != 0)
        assert np.all(~np.isnan(wx))

    def test_biased_Z_noise_different_weights(self):
        L = 10
        p_X, p_Y, p_Z = 0, 0, 0.4
        wx, wz = self.get_weights(p_X, p_Y, p_Z, L)
        assert np.any(wx != 0)
        assert np.any(wx != wx[0])

    def test_only_x_edges_different_wx(self):
        L = 10
        p_X, p_Y, p_Z = 0.3, 0.1, 0
        wx, wz = self.get_weights(p_X, p_Y, p_Z, L)
        assert np.any(wx != wx[0])
        code = Toric3DCode(L, L, L)
        x_edge_indices = [
            index for index, edge in enumerate(code.qubit_index)
            if code.qubit_axis(edge) == 'x'
        ]
        y_edge_indices = [
            index for index, edge in enumerate(code.qubit_index)
            if code.qubit_axis(edge) == 'y'
        ]
        z_edge_indices = [
            index for index, edge in enumerate(code.qubit_index)
            if code.qubit_axis(edge) == 'z'
        ]

        # Weights for the same wedge type should be equal.
        assert np.all(wx[x_edge_indices] == wx[x_edge_indices[0]])
        assert np.all(wx[y_edge_indices] == wx[y_edge_indices[0]])
        assert np.all(wx[z_edge_indices] == wx[z_edge_indices[0]])

        # Weights for y-edges and z-edges should be equal.
        assert np.all(wx[x_edge_indices] == wx[y_edge_indices])

        # Weights for x-edges and z-edges should not be equal.
        assert np.any(wx[x_edge_indices] != wx[z_edge_indices])


@pytest.mark.parametrize('pauli,bias,expected', [
    ('X', 0.5, (1/3, 1/3, 1/3)),
    ('Y', 0.5, (1/3, 1/3, 1/3)),
    ('Z', 0.5, (1/3, 1/3, 1/3)),
    ('X', np.inf, (1, 0, 0)),
    ('Y', np.inf, (0, 1, 0)),
    ('Z', np.inf, (0, 0, 1)),
    ('X', 1, (0.5, 0.25, 0.25)),
    ('Y', 1, (0.25, 0.5, 0.25)),
    ('Z', 1, (0.25, 0.25, 0.5)),
])
def test_get_direction_from_bias_ratio(pauli, bias, expected):
    expected_params = dict(zip(['r_x', 'r_y', 'r_z'], expected))
    params = get_direction_from_bias_ratio(pauli, bias)
    for key in expected_params.keys():
        assert np.isclose(params[key], expected_params[key])



import pytest
import numpy as np
from panqec.codes import Toric3DCode, StabilizerCode
from panqec.error_models import PauliErrorModel
from panqec.decoders import (
    SweepDecoder3D, SweepMatchDecoder, MatchingDecoder
)


@pytest.fixture
def code():
    return Toric3DCode(3, 4, 5)


@pytest.fixture
def rng():
    np.random.seed(0)
    return np.random


def get_pymatching_distance_matrix(matching):
    edges = matching.edges()
    nodes = sorted(
        set([edge[0] for edge in edges] + [edge[1] for edge in edges])
    )
    weights = {
        (edge[0], edge[1]): edge[2]['weight']
        for edge in edges
    }
    matrix = np.zeros((len(nodes), len(nodes)), dtype=float)
    for i_1, node_1 in enumerate(nodes):
        for i_2, node_2 in enumerate(nodes):
            if (node_1, node_2) in weights:
                matrix[i_1, i_2] = weights[(i_1, i_2)]
    return matrix


class TestDeformedXZZXErrorModel:

    @pytest.mark.parametrize(
        'noise, original, deformed',
        [
            ((1, 0, 0), 'X', 'Z'),
            ((0, 0, 1), 'Z', 'X'),
            ((0, 1, 0), 'Y', 'Y'),
        ]
    )
    def test_max_noise(self, code, rng, noise, original, deformed):
        error_model = PauliErrorModel(
            *noise, deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error = error_model.generate(code, error_rate=1, rng=rng)
        pauli = code.from_bsf(error)
        for edge in code.qubit_coordinates:
            if code.qubit_axis(edge) == 'z':
                assert pauli[edge] == deformed
            else:
                assert pauli[edge] == original

    def test_original_all_X_becomes_Z_on_deformed_axis(self, code):
        error_model = PauliErrorModel(
            1, 0, 0,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error = error_model.generate(code, error_rate=1)
        pauli = code.from_bsf(error)
        print(pauli)

        for edge in code.qubit_index:
            if code.qubit_axis(edge) == 'z':
                assert pauli[edge] == 'Z'
            else:
                assert pauli[edge] == 'X'

    def test_original_all_Z_becomes_X_on_deformed_axis(self, code):
        error_model = PauliErrorModel(
            0, 0, 1,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error = error_model.generate(code, error_rate=1)
        pauli = code.from_bsf(error)

        for edge in code.qubit_index:
            if code.qubit_axis(edge) == 'z':
                assert pauli[edge] == 'X'
            else:
                assert pauli[edge] == 'Z'

    def test_all_Y_deformed_is_still_all_Y(self, code):
        error_model = PauliErrorModel(
            0, 1, 0,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error = error_model.generate(code, error_rate=1)
        pauli = code.from_bsf(error)

        for edge in code.qubit_index:
            assert pauli[edge] == 'Y'

    def test_label(self, code):
        error_model = PauliErrorModel(
            1, 0, 0,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        assert error_model.label == 'Deformed XZZX Pauli X1.0000Y0.0000Z0.0000'

    def test_decode_trivial(self, code):
        error_model = PauliErrorModel(
            0.1, 0.2, 0.7,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.1
        decoder = SweepMatchDecoder(code, error_model, error_rate)

        syndrome = np.zeros(code.stabilizer_matrix.shape[0], dtype=np.uint)
        correction = decoder.decode(syndrome)
        assert np.all(correction == 0)
        assert issubclass(correction.dtype.type, np.integer)

    def test_decode_single_X_on_undeformed_axis(self, code):
        error_model = PauliErrorModel(
            0.1, 0.2, 0.7,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.1
        decoder = SweepMatchDecoder(code, error_model, error_rate)

        # Single-qubit X error on undeformed edge.
        error = code.to_bsf({
            (0, 1, 0): 'X'
        })
        assert np.any(error != 0)

        # Calculate the syndrome and make sure it's nontrivial.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        # Total error should be in code space.
        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    @pytest.mark.parametrize(
        'operator, location',
        [
            ['X', (0, 0, 0, 0)],
            ['Y', (0, 1, 0, 0)],
            ['Z', (0, 0, 1, 0)],
            ['X', (1, 0, 0, 1)],
            ['Y', (1, 0, 2, 0)],
            ['Z', (1, 2, 0, 0)],
            ['X', (2, 0, 0, 2)],
            ['Y', (2, 1, 1, 0)],
            ['Z', (2, 0, 2, 0)],
        ]
    )
    def test_decode_single_qubit_error(
        self, code, operator, location
    ):
        noise_direction = (0.1, 0.2, 0.7)
        error_model = PauliErrorModel(
            *noise_direction,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.1
        decoder = SweepMatchDecoder(code, error_model, error_rate)

        # Single-qubit X error on undeformed edge.
        error = code.to_bsf({
            (0, 1, 0): 'X'
        })
        assert np.any(error != 0)

        # Calculate the syndrome and make sure it's nontrivial.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        # Total error should be in code space.
        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    # TODO fix pymatching
    @pytest.mark.skip
    def test_deformed_pymatching_weights_nonuniform(self, code):
        error_model = PauliErrorModel(
            0.1, 0.2, 0.7,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.1
        decoder = SweepMatchDecoder(code, error_model, error_rate)
        assert decoder.matcher.error_model.direction == (0.1, 0.2, 0.7)
        matching = decoder.matcher.matcher_x
        distance_matrix = get_pymatching_distance_matrix(matching)
        n_vertices = int(np.product(code.size))
        assert distance_matrix.shape == (n_vertices, n_vertices)

        # The index of the origin vertex.
        origin_index = [
            index
            for index, location in enumerate(code.stabilizer_coordinates)
            if location == (0, 0, 0)
            and code.stabilizer_type(location) == 'vertex'
        ][0]

        # Distances from the origin vertex.
        origin_distances = np.zeros(code.size)

        for index, coordinate in enumerate(code.stabilizer_index):
            if code.stabilizer_type(coordinate) == 'vertex':
                location = tuple(
                    (np.array(coordinate)/2).astype(int).tolist()
                )
                origin_distances[location] = distance_matrix[
                    origin_index, index
                ]

        assert origin_distances[0, 0, 0] == 0

        # Distances in the undeformed direction should be equal.
        assert origin_distances[1, 0, 0] == origin_distances[0, 1, 0]

        # Distances in the deformed direction should be different.
        assert origin_distances[0, 1, 0] != origin_distances[0, 0, 1]

    # TODO fix pymatching new version
    @pytest.mark.skip
    def test_equal_XZ_bias_deformed_pymatching_weights_uniform(self, code):
        error_model = PauliErrorModel(
            0.4, 0.2, 0.4,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        print(f'{error_model.direction=}')
        error_rate = 0.1
        decoder = SweepMatchDecoder(code, error_model, error_rate)
        assert decoder.matcher.error_model.direction == (0.4, 0.2, 0.4)
        matching = decoder.matcher.matcher_x
        distance_matrix = get_pymatching_distance_matrix(matching)
        n_vertices = int(np.product(code.size))
        assert distance_matrix.shape == (n_vertices, n_vertices)

        # Distances from the origin vertex.
        origin_distances = distance_matrix[0].reshape(code.size)
        assert origin_distances[0, 0, 0] == 0

        # Distances in the undeformed direction should be equal.
        assert origin_distances[0, 1, 0] == origin_distances[0, 0, 1]

        # Distances in the deformed direction should be different.
        assert origin_distances[1, 0, 0] == origin_distances[0, 0, 1]


class TestDeformedSweepDecoder3D:

    def test_decode_trivial(self, code):
        error_model = PauliErrorModel(
            1/3, 1/3, 1/3,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.5
        decoder = SweepDecoder3D(code, error_model, error_rate)
        n = code.n
        error = np.zeros(2*n, dtype=np.uint)
        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)
        total_error = (correction + error) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)
        assert issubclass(correction.dtype.type, np.integer)

    def test_all_3_faces_active(self, code):
        error_pauli = dict()
        sites = [
            (3, 2, 2), (2, 4, 3)
        ]
        for site in sites:
            error_pauli[site] = 'Z'
        error = code.to_bsf(error_pauli)
        error_model = PauliErrorModel(
            1/3, 1/3, 1/3,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.5
        decoder = SweepDecoder3D(code, error_model, error_rate)
        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)


class TestMatchingDecoder:

    def test_decode_trivial(self, code):
        error_model = PauliErrorModel(
            1/3, 1/3, 1/3,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.5
        decoder = MatchingDecoder(code, error_model, error_rate,
                                  error_type='X')
        n = code.n
        error = np.zeros(2*n, dtype=np.uint)
        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)
        total_error = (correction + error) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)
        assert issubclass(correction.dtype.type, np.integer)


class XNoiseOnYZEdgesOnly(PauliErrorModel):
    """X noise applied on y and z edges only."""

    def __init__(self):
        super(XNoiseOnYZEdgesOnly, self).__init__(1, 0, 0)

    def generate(
        self, code: StabilizerCode, error_rate: float, rng=None
    ) -> np.ndarray:
        error = super(XNoiseOnYZEdgesOnly, self).generate(
            code, error_rate, rng=rng
        )
        for index, location in enumerate(code.qubit_coordinates):
            if code.qubit_axis(location) == 'x':
                error[index] = 0
        return error


class TestMatchingXNoiseOnYZEdgesOnly:

    def test_decode(self, code):
        for seed in range(5):
            rng = np.random.default_rng(seed=seed)
            error_rate = 0.5
            error_model = XNoiseOnYZEdgesOnly()
            decoder = MatchingDecoder(
                code, error_model, error_rate, error_type='X'
            )
            error = error_model.generate(
                code, error_rate=error_rate, rng=rng
            )
            assert any(error), 'Error should be non-trivial'
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            assert any(correction), 'Correction should be non-trivial'
            total_error = (correction + error) % 2
            assert np.all(
                code.measure_syndrome(total_error) == 0
            ), 'Total error should be in code space'

            # Error and correction as objects.
            error_pauli = code.from_bsf(error)
            correction_pauli = code.from_bsf(correction)

            x_edges = [
                edge for edge in code.qubit_index
                if code.qubit_axis(edge) == 'x'
            ]
            y_edges = [
                edge for edge in code.qubit_index
                if code.qubit_axis(edge) == 'y'
            ]
            z_edges = [
                edge for edge in code.qubit_index
                if code.qubit_axis(edge) == 'z'
            ]

            assert np.all(
                edge not in error_pauli
                or error_pauli[edge] == 'I'
                for edge in x_edges
            ), 'No errors should be on x edges'

            assert np.all(
                edge not in correction_pauli
                or correction_pauli[edge] == 'I'
                for edge in x_edges
            ), 'No corrections should be on x edges'

            assert np.any([
                correction_pauli[edge] != 'I'
                for edge in y_edges + z_edges
                if edge in correction_pauli
            ]), 'Non-trivial corrections should be on the y and z edges'



import pytest
from panqec.codes import Toric2DCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestToric2DCode(StabilizerCodeTest):

    @pytest.fixture(params=[(2, 2), (3, 3), (2, 3)])
    def code(self, request):
        return Toric2DCode(*request.param)



import numpy as np
from panqec.codes import StabilizerCode
from panqec.decoders import (
    BaseDecoder, SweepDecoder3D, MatchingDecoder
)
from panqec.error_models import BaseErrorModel


class SweepMatchDecoder(BaseDecoder):

    label = 'Toric 3D Sweep + Matching Decoder'
    allowed_codes = ["Toric3DCode", "Planar3DCode"]

    sweeper: BaseDecoder
    matcher: BaseDecoder

    def __init__(self, code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float):
        super().__init__(code, error_model, error_rate)
        self.sweeper = SweepDecoder3D(code, error_model, error_rate)
        self.matcher = MatchingDecoder(code, error_model, error_rate,
                                       error_type='X')

    @property
    def params(self) -> dict:
        return {}

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X and Z corrections given code and measured syndrome."""

        z_correction = self.sweeper.decode(syndrome)
        x_correction = self.matcher.decode(syndrome)

        correction = (x_correction + z_correction) % 2

        return correction
